# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Server-side task to mirror an external APT suite into a debusine."""

import binascii
import functools
import logging
import os
import shlex
import shutil
import subprocess
import tempfile
from collections import defaultdict
from collections.abc import Generator, Sequence
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field
from enum import Enum, auto
from operator import itemgetter
from pathlib import Path, PurePath
from textwrap import dedent
from typing import TYPE_CHECKING, override
from urllib.parse import ParseResult, urlparse, urlunparse

import requests
from debian.deb822 import Deb822, Packages, Release, Sources
from django.db import transaction
from django_pglocks import advisory_lock

from debusine.artifacts.local_artifact import (
    BinaryPackage,
    RepositoryIndex,
    SourcePackage,
)
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    WorkRequestResults,
)
from debusine.assets.models import AssetCategory, BasicAPTAuthenticationData
from debusine.client.client_utils import download_file
from debusine.db.locks import LockError, LockType
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Asset,
    AssetUsage,
    CollectionItem,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.db.tasks.inputs import CollectionInput
from debusine.server.collections import DebianSuiteManager
from debusine.server.tasks import BaseServerTask
from debusine.server.tasks.models import APTMirrorData
from debusine.utils import calculate_hash
from debusine.utils.urls import URLPath

if TYPE_CHECKING:
    from requests.sessions import _Auth


@dataclass(frozen=True, kw_only=True)
class IndexFile:
    """Details of an index file found in a ``Release`` file."""

    component: str
    architecture: str | None = None
    size: int
    sha256: str
    must_exist: bool


class IndexFiles(dict[str, IndexFile]):
    """Alternatives for a single index file, indexed by compression suffix."""


@dataclass(frozen=True, kw_only=True, slots=True)
class IndexedPackage[T: Deb822]:
    """Details of a source or binary package found in an index."""

    contents: T
    component: str
    ignore_unauthorized: bool


class FilterMatchResult(Enum):
    """Results of matching a package against a list of filters."""

    #: No filter matched this package.
    NO_MATCH = auto()

    #: A filter matched this package.
    MATCH = auto()

    #: A filter matched this package.  When downloading it, the task should
    #: ignore failures indicating insufficient authorization (HTTP 401 or
    #: 403).
    MATCH_IGNORE_UNAUTHORIZED = auto()

    @property
    def matched(self) -> bool:
        """Return whether a filter matched this package."""
        return self in {
            FilterMatchResult.MATCH,
            FilterMatchResult.MATCH_IGNORE_UNAUTHORIZED,
        }


@dataclass(frozen=True, kw_only=True, slots=True)
class PlanAdd[T: Deb822]:
    """A plan for adding a single item to a collection."""

    name: str
    contents: T
    component: str
    ignore_unauthorized: bool = False


@dataclass(frozen=True, kw_only=True, slots=True)
class PlanReplace[T: Deb822](PlanAdd[T]):
    """A plan for replacing a single item in a collection."""

    item: CollectionItem


@dataclass(frozen=True, kw_only=True, slots=True)
class Plan[T: Deb822]:
    """A plan for updating items in a collection."""

    add: list[PlanAdd[T]] = field(default_factory=list)
    replace: list[PlanReplace[T]] = field(default_factory=list)
    remove: list[CollectionItem] = field(default_factory=list)


class InconsistentMirrorError(Exception):
    """The remote mirror is inconsistent."""


logger = logging.getLogger(__name__)


def _run_and_log_errors(
    args: Sequence[str],
    cwd: Path | None = None,
    env: dict[str, str] | None = None,
) -> subprocess.CompletedProcess[str]:
    """Run a subprocess, logging stderr on failure."""
    try:
        return subprocess.run(
            args, cwd=cwd, env=env, text=True, check=True, capture_output=True
        )
    except subprocess.CalledProcessError as e:
        logger.error("Error output from %s:\n%s", shlex.join(e.cmd), e.stderr)
        raise


class APTMirror(BaseServerTask[APTMirrorData]):
    """Task that mirrors an external APT suite into a debusine collection."""

    TASK_VERSION = 1
    TASK_MANAGES_TRANSACTIONS = True

    collection = CollectionInput(category=CollectionCategory.SUITE)

    @functools.cached_property
    def collection_manager(self) -> DebianSuiteManager:
        """The manager for the suite this task is updating."""
        manager = self.collection.collection.manager
        assert isinstance(manager, DebianSuiteManager)
        return manager

    @functools.cached_property
    def authentication_data(self) -> BasicAPTAuthenticationData | None:
        """Authentication data to use for this mirroring task, if any."""
        if self.data.authentication is None:
            return None

        authentication = Asset.objects.get_by_slug(
            category=AssetCategory.APT_AUTHENTICATION,
            slug=self.data.authentication,
            workspace=self.workspace,
        )
        try:
            asset_usage = authentication.usage.get(workspace=self.workspace)
        except AssetUsage.DoesNotExist:
            has_permission = False
        else:
            has_permission = asset_usage.can_use_apt_authentication_with(
                PermissionContext.for_work_request(
                    work_request=self.work_request
                )
            )
        if not has_permission:
            raise InconsistentMirrorError(
                f"{self.work_request.created_by} cannot use asset"
                f" {self.data.authentication!r} for APT authentication"
            )
        assert isinstance(authentication.data_model, BasicAPTAuthenticationData)
        return authentication.data_model

    def make_apt_environment(self, temp_path: Path) -> dict[str, str]:
        """Make a suitable process environment for running apt commands."""
        env = os.environ.copy()
        env["APT_CONFIG"] = str(temp_path / "etc/apt/apt.conf")
        return env

    def write_apt_config(
        self, path: Path, options: list[tuple[str, str | None]]
    ) -> None:
        """Write APT configuration options out to a file."""
        apt_config = [
            f'#clear {key};' if value is None else f'{key} "{value}";'
            for key, value in options
        ]
        path.write_text("\n".join(apt_config) + "\n")

    def get_apt_lists_path(self, temp_path: Path) -> Path:
        """Get the path where APT stores downloaded index files."""
        # We have to import this late, as otherwise mypy_django_plugin tries
        # to import it during early initialization and fails (since we can't
        # install it from PyPI and so it isn't present in mypy's virtual
        # environment).
        import apt_pkg

        with self.apt_config(temp_path):
            return Path(apt_pkg.config.find_dir("Dir::State::lists"))

    @functools.cache
    def get_apt_base_uri(self, temp_path: Path) -> str:
        """Get APT's version of the base repository download URI."""
        # We have to import this late, as otherwise mypy_django_plugin tries
        # to import it during early initialization and fails (since we can't
        # install it from PyPI and so it isn't present in mypy's virtual
        # environment).
        import apt_pkg

        with self.apt_config(temp_path):
            source_list = apt_pkg.SourceList()
            source_list.read_main_list()
            assert len(source_list.list) == 1
            assert isinstance(source_list.list[0].uri, str)
            return source_list.list[0].uri

    @property
    def suite_prefix(self) -> str:
        """Return the URL prefix for the suite."""
        if self.data.suite.endswith("/"):
            return self.data.suite
        else:
            return f"dists/{self.data.suite}/"

    def get_apt_release_file_name(
        self, temp_path: Path, release_name: str
    ) -> str:
        """Get the file name where APT stores a given ``Release`` file."""
        # We have to import this late, as otherwise mypy_django_plugin tries
        # to import it during early initialization and fails (since we can't
        # install it from PyPI and so it isn't present in mypy's virtual
        # environment).
        import apt_pkg

        name = apt_pkg.uri_to_filename(
            self.get_apt_base_uri(temp_path) + self.suite_prefix + release_name
        )
        assert isinstance(name, str)
        return name

    def fetch_meta_indexes(self, temp_path: Path) -> None:
        """Fetch meta-indexes (``Release`` files) for this suite."""
        (temp_path / "etc/apt/apt.conf.d").mkdir(parents=True)
        (temp_path / "etc/apt/preferences.d").mkdir(parents=True)
        (temp_path / "etc/apt/sources.list.d").mkdir(parents=True)
        (temp_path / "var/lib/apt/lists/partial").mkdir(parents=True)

        apt_config: list[tuple[str, str | None]] = [
            ("Dir", str(temp_path)),
            ("Acquire::IndexTargets", None),
        ]
        apt_config_path = temp_path / "etc/apt/apt.conf"
        self.write_apt_config(apt_config_path, apt_config)

        if self.authentication_data is not None:
            apt_auth_path = (
                temp_path
                / f"etc/apt/auth.conf.d/{self.data.authentication}.conf"
            )
            apt_auth_path.parent.mkdir(parents=True)
            apt_auth_path.write_text(
                dedent(
                    f"""\
                    machine {self.data.url}
                    login {self.authentication_data.credentials.username}
                    password {self.authentication_data.credentials.password}
                    """
                )
            )

        source = {
            "Types": "deb deb-src",
            "URIs": str(self.data.url),
            "Suites": self.data.suite,
        }
        if self.data.components is not None:
            source["Components"] = " ".join(self.data.components)
        if self.data.signing_key is not None:
            source["Signed-By"] = "\n" + "\n".join(
                f" {line}" if line else " ."
                for line in self.data.signing_key.splitlines()
            )
        (temp_path / "etc/apt/sources.list.d/mirror.sources").write_text(
            str(Deb822(source))
        )

        # Since we clear Acquire::IndexTargets, this just fetches and
        # verifies the Release files.
        logger.info(
            "Fetching meta-indexes for %s %s", self.data.url, self.data.suite
        )
        _run_and_log_errors(
            ["apt-get", "update"], env=self.make_apt_environment(temp_path)
        )

    def get_release_path(self, temp_path: Path) -> Path:
        """Get the preferred path to the fetched ``Release`` file."""
        for release_name in ("InRelease", "Release"):
            lists_path = self.get_apt_lists_path(temp_path)
            release_path = lists_path / self.get_apt_release_file_name(
                temp_path, release_name
            )
            if release_path.exists():
                return release_path
        raise InconsistentMirrorError(
            "Cannot mirror a repository without a Release/InRelease file"
        )

    def get_index_files(
        self, temp_path: Path, only_type: str | None = None
    ) -> dict[str, IndexFiles]:
        """Find all relevant index files mentioned by the ``Release`` file."""
        with self.get_release_path(temp_path).open(mode="rb") as release_file:
            release = Release(release_file)

        # An index file may be stored in multiple compression formats; from
        # our point of view, fetching any single one of them will do.
        # Uncompressed indexes may have checksums in the Release file
        # without actually existing on the server.
        # https://wiki.debian.org/DebianRepository/Format#Compression_of_indices
        # https://wiki.debian.org/DebianRepository/Format#MD5Sum.2C_SHA1.2C_SHA256
        index_files: dict[str, IndexFiles] = defaultdict(IndexFiles)
        for entry in release["SHA256"]:
            entry_path = PurePath(entry["name"])
            index_type = entry_path.stem.split("-", 1)[0]
            if only_type not in {None, index_type}:
                continue

            if index_type in {
                "Contents",
                "Packages",
                "Sources",
                "Translation",
            }:
                if len(entry_path.parts) > 1:
                    component = entry_path.parts[0]
                else:
                    component = "main"
            else:
                continue

            architecture: str | None = None
            if (
                index_type == "Packages"
                and len(entry_path.parts) > 1
                and entry_path.parts[-2].startswith("binary-")
            ):
                architecture = entry_path.parts[-2].removeprefix("binary-")
            elif entry_path.stem.startswith("Contents-"):
                architecture = entry_path.stem.rsplit("-", 1)[1]

            base_rel_path = str(entry_path.parent / entry_path.stem)
            index_files[base_rel_path][entry_path.suffix or ""] = IndexFile(
                component=component,
                architecture=architecture,
                size=int(entry["size"]),
                sha256=entry["sha256"],
                must_exist=bool(entry_path.suffix),
            )
        return index_files

    @functools.cached_property
    def base_url_parsed(self) -> ParseResult:
        """A parsed version of the archive's base URL."""
        return urlparse(str(self.data.url))

    def fetch_archive_file(
        self,
        root: Path,
        trailing_path: str,
        *,
        expected_size: int,
        expected_sha256: str,
        must_exist: bool = True,
        ignore_unauthorized: bool = False,
    ) -> bool:
        """
        Fetch a URL from the archive given the trailing part of its path.

        :param root: the root of the directory structure where files should
          be fetched.
        :param trailing_path: the trailing part of the URL path, appended to
          each of the archive's base URL and the ``root`` parameter.
        :param expected_size: the expected size of the fetched file.
        :param expected_sha256: the expected SHA256 hex digest of the
          fetched file.
        :param must_exist: if True (default), fail if fetching the URL
          returns HTTP 404; otherwise, raise an exception in such cases.
        :param ignore_unauthorized: if True, return False if fetching the
          URL returns HTTP 401 or 403; otherwise, raise an exception in such
          cases.
        :raises requests.HTTPError: if fetching the URL failed.
        :raises InconsistentMirrorError: if the contents of the URL do not
          have the expected size and SHA256 digest.
        :return: True if the URL was successfully fetched; False if fetching
          the URL failed but the error was ignored due to
          ``must_exist=False`` or ``ignore_unauthorized=True``.
        """
        url = urlunparse(
            self.base_url_parsed._replace(
                path=str(
                    URLPath(self.base_url_parsed.path, is_file=False)
                    + trailing_path
                )
            )
        )
        target_path = root / trailing_path
        if not target_path.resolve().is_relative_to(root):
            raise InconsistentMirrorError(f"{trailing_path!r} escapes {root}")
        target_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            auth: _Auth | None = None
            if self.authentication_data:
                auth = (
                    self.authentication_data.credentials.username,
                    self.authentication_data.credentials.password,
                )
            stats = download_file(
                url, target_path, auth=auth, hashes=("sha256",)
            )
        except requests.HTTPError as e:
            if (
                isinstance(e.response, requests.Response)
                and e.response.status_code == requests.codes.not_found
                and not must_exist
            ):
                logger.info("%s not found; skipping", url)
                return False
            elif (
                isinstance(e.response, requests.Response)
                and e.response.status_code
                in {requests.codes.unauthorized, requests.codes.forbidden}
                and ignore_unauthorized
            ):
                logger.info("%s not authorized; skipping", url)
                return False
            logger.exception("Failed to download %s", url)
            raise
        else:
            assert isinstance(stats["size"], int)
            if stats["size"] != expected_size:
                target_path.unlink()
                raise InconsistentMirrorError(
                    f"Expected {trailing_path!r} to have size "
                    f"{expected_size} bytes; got {stats['size']} bytes"
                )
            assert isinstance(stats["sha256"], str)
            if stats["sha256"] != expected_sha256:
                target_path.unlink()
                raise InconsistentMirrorError(
                    f"Expected {trailing_path!r} to have SHA256 "
                    f"{expected_sha256!r}; got {stats['sha256']!r}"
                )
            return True

    def fetch_indexes(self, temp_path: Path) -> None:
        """Fetch indexes for this suite."""
        (indexes_path := temp_path / "indexes").mkdir()
        compress_exts = (".xz", ".bz2", ".gz", "")
        architectures: set[str] | None = None
        if self.data.architectures is not None:
            architectures = set(self.data.architectures) | {"all"}
        for rel_path, index_files in sorted(
            self.get_index_files(temp_path).items()
        ):
            for compress_ext in compress_exts:
                if (index_file := index_files.get(compress_ext)) is None:
                    continue
                if (
                    self.data.components is not None
                    and index_file.component not in self.data.components
                ):
                    logger.info(
                        "Skipping %s (component not in %s)",
                        rel_path + compress_ext,
                        self.data.components,
                    )
                    break
                if (
                    architectures is not None
                    and index_file.architecture is not None
                    and index_file.architecture not in architectures
                ):
                    logger.info(
                        "Skipping %s (architecture not in %s)",
                        rel_path + compress_ext,
                        sorted(architectures),
                    )
                    break

                if self.fetch_archive_file(
                    indexes_path,
                    self.suite_prefix + rel_path + compress_ext,
                    expected_size=index_file.size,
                    expected_sha256=index_file.sha256,
                    must_exist=index_file.must_exist,
                ):
                    break
            else:
                raise InconsistentMirrorError(
                    f"No alternatives found for {rel_path!r}"
                )

    def get_fetched_index_files(
        self, temp_path: Path, only_type: str | None = None
    ) -> dict[str, tuple[Path, IndexFile]]:
        """Find index files that were successfully fetched."""
        indexes_path = temp_path / "indexes"
        indexes: dict[str, tuple[Path, IndexFile]] = {}
        for base_rel_path, index_files in self.get_index_files(
            temp_path, only_type=only_type
        ).items():
            for suffix, index_file in index_files.items():
                rel_path = base_rel_path + suffix
                index_path = indexes_path / (self.suite_prefix + rel_path)
                if index_path.exists():
                    indexes[rel_path] = (index_path, index_file)
        return indexes

    def match_filters(
        self,
        *,
        source_name: str,
        binary_name: str | None = None,
        priority: str | None = None,
        section: str | None = None,
    ) -> FilterMatchResult:
        """Check whether a package matches any of the configured filters."""
        if self.data.filters is None:
            return FilterMatchResult.MATCH

        result = FilterMatchResult.NO_MATCH
        for filt in self.data.filters:
            if (
                (
                    filt.binary_name is None
                    or (
                        binary_name is not None
                        and filt.binary_name.match(binary_name)
                    )
                )
                and (filt.priority is None or filt.priority == priority)
                and (filt.section is None or filt.section == section)
                and (
                    filt.source_name is None
                    or filt.source_name.match(source_name)
                )
            ):
                if filt.ignore_unauthorized:
                    result = FilterMatchResult.MATCH_IGNORE_UNAUTHORIZED
                    break
                else:
                    result = FilterMatchResult.MATCH
        return result

    def plan_sources(self, temp_path: Path) -> Plan[Sources]:
        """Plan the update of all source packages in the collection."""
        # Source packages from the remote collection
        indexes: dict[str, IndexedPackage[Sources]] = {}
        for index_path, index_file in self.get_fetched_index_files(
            temp_path, only_type="Sources"
        ).values():
            for source in Sources.iter_paragraphs(index_path, use_apt_pkg=True):
                name = "{Package}_{Version}".format(**source)
                if name in indexes:
                    raise InconsistentMirrorError(
                        f"{name} found in multiple components: "
                        f"{indexes[name].component} and {index_file.component}"
                    )

                match_result = self.match_filters(
                    source_name=source["Package"],
                    priority=source.get("Priority"),
                    section=source.get("Section"),
                )
                if match_result.matched:
                    indexes[name] = IndexedPackage[Sources](
                        contents=source,
                        component=index_file.component,
                        ignore_unauthorized=(
                            match_result
                            == FilterMatchResult.MATCH_IGNORE_UNAUTHORIZED
                        ),
                    )

        # Source packages from the local collection
        items: dict[str, tuple[CollectionItem, dict[str, str]]] = {}
        for item in (
            self.collection.collection.child_items.active()
            .filter(
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.SOURCE_PACKAGE,
            )
            .only("name", "data", "artifact")
            .prefetch_related("artifact__fileinartifact_set__file")
        ):
            artifact = item.artifact
            assert artifact is not None
            items[item.name] = (
                item,
                {
                    file_in_artifact.path: file_in_artifact.file.sha256.hex()
                    for file_in_artifact in artifact.fileinartifact_set.all()
                },
            )

        plan = Plan[Sources]()

        for name, indexed in sorted(indexes.items()):
            index_checksums = {
                checksum["name"]: checksum["sha256"]
                for checksum in indexed.contents["Checksums-Sha256"]
            }
            if name in items:
                item, item_checksums = items[name]
                if index_checksums != item_checksums:
                    plan.replace.append(
                        PlanReplace[Sources](
                            name=name,
                            contents=indexed.contents,
                            component=indexed.component,
                            item=item,
                            ignore_unauthorized=indexed.ignore_unauthorized,
                        )
                    )
            else:
                plan.add.append(
                    PlanAdd[Sources](
                        name=name,
                        contents=indexed.contents,
                        component=indexed.component,
                        ignore_unauthorized=indexed.ignore_unauthorized,
                    )
                )

        for name in sorted(items.keys() - indexes.keys()):
            item, _ = items[name]
            plan.remove.append(item)

        return plan

    def add_source(self, temp_path: Path, plan: PlanAdd[Sources]) -> None:
        """Download a source package and add it to the collection."""
        with ExitStack() as stack:
            (temp_download := temp_path / "download").mkdir(parents=True)
            stack.callback(shutil.rmtree, temp_download)

            source_artifact: Artifact | None = None
            if (
                self.collection_manager.parent_archive is not None
                and (
                    source_from_archive
                    := self.collection_manager.parent_archive.manager.lookup(
                        f"source-version:{plan.name}"
                    )
                )
                is not None
            ):
                # The parent archive already has this package, so we don't
                # need to download it again as long as its checksums match.
                assert source_from_archive.artifact is not None
                if sorted(
                    source_from_archive.artifact.fileinartifact_set.values_list(
                        "path", "file__size", "file__sha256"
                    ),
                    key=itemgetter(0),
                ) == sorted(
                    [
                        (
                            checksum["name"],
                            int(checksum["size"]),
                            binascii.unhexlify(checksum["sha256"]),
                        )
                        for checksum in plan.contents["Checksums-Sha256"]
                    ],
                    key=itemgetter(0),
                ):
                    source_artifact = source_from_archive.artifact

            if source_artifact is None:
                # We don't have this package anywhere usable, so we need to
                # download it again.
                # https://wiki.debian.org/DebianRepository/Format#A.22Sources.22_Indices
                # says this is mandatory, but "apt-ftparchive sources" omits
                # it if it's "./".
                directory = PurePath(plan.contents.get("Directory", "."))
                for checksum in plan.contents["Checksums-Sha256"]:
                    if not self.fetch_archive_file(
                        temp_download,
                        (directory / checksum["name"]).as_posix(),
                        expected_size=int(checksum["size"]),
                        expected_sha256=checksum["sha256"],
                        ignore_unauthorized=plan.ignore_unauthorized,
                    ):
                        # Unauthorized, but ignored by filters.
                        return
                source_package = SourcePackage.create(
                    name=plan.contents["Package"],
                    version=plan.contents["Version"],
                    files=list((temp_download / directory).iterdir()),
                )
                source_artifact = Artifact.objects.create_from_local_artifact(
                    source_package,
                    self.workspace,
                    created_by_work_request=self.work_request,
                )

            with self.collection_manager.allow_rewinding_versions():
                self.collection_manager.add_artifact(
                    source_artifact,
                    user=self.work_request.created_by,
                    variables={
                        "component": plan.component,
                        "section": plan.contents["Section"],
                    },
                    replace=self.collection_manager.may_reuse_versions,
                )

    def update_sources(self, temp_path: Path, plan: Plan[Sources]) -> None:
        """
        Update all source packages in the collection.

        This may take a long time, so it commits transactions as it goes
        rather than taking a single long transaction.  If processing an item
        fails, then the results of earlier processing will remain visible.
        """
        for add in plan.add:
            with transaction.atomic():
                self.add_source(temp_path, add)

        for replace in plan.replace:
            with transaction.atomic():
                self.collection_manager.remove_item(replace.item)
                self.add_source(temp_path, replace)

        with transaction.atomic():
            for remove in plan.remove:
                self.collection_manager.remove_item(remove)

    def plan_binaries(self, temp_path: Path) -> Plan[Packages]:
        """Plan the update of all binary packages in the collection."""
        # Binary packages from the remote collection
        indexes: dict[str, IndexedPackage[Packages]] = {}
        for index_path, index_file in self.get_fetched_index_files(
            temp_path, only_type="Packages"
        ).values():
            for binary in Packages.iter_paragraphs(
                index_path, use_apt_pkg=True
            ):
                name = "{Package}_{Version}_{Architecture}".format(**binary)
                if name in indexes:
                    if indexes[name].component != index_file.component:
                        raise InconsistentMirrorError(
                            f"{name} found in multiple components: "
                            f"{indexes[name].component} and "
                            f"{index_file.component}"
                        )
                    elif indexes[name].contents != binary:
                        raise InconsistentMirrorError(
                            f"{name} mismatch.  Conflicting Packages "
                            f"entries:\n\n"
                            + indexes[name].contents.dump()
                            + "\n\n"
                            + binary.dump()
                        )

                match_result = self.match_filters(
                    source_name=binary.get("Source", binary["Package"]).split()[
                        0
                    ],
                    binary_name=binary["Package"],
                    priority=binary.get("Priority"),
                    section=binary.get("Section"),
                )
                if match_result.matched:
                    indexes[name] = IndexedPackage[Packages](
                        contents=binary,
                        component=index_file.component,
                        ignore_unauthorized=(
                            match_result
                            == FilterMatchResult.MATCH_IGNORE_UNAUTHORIZED
                        ),
                    )

        # Binary packages from the local collection
        items: dict[str, tuple[CollectionItem, dict[str, str]]] = {}
        for item in (
            self.collection.collection.child_items.active()
            .filter(
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.BINARY_PACKAGE,
            )
            .only("name", "data", "artifact")
            .prefetch_related("artifact__fileinartifact_set__file")
        ):
            artifact = item.artifact
            assert artifact is not None
            items[item.name] = (
                item,
                {
                    file_in_artifact.path: file_in_artifact.file.sha256.hex()
                    for file_in_artifact in artifact.fileinartifact_set.all()
                },
            )

        plan = Plan[Packages]()

        for name, indexed in sorted(indexes.items()):
            index_checksums = {
                PurePath(indexed.contents["Filename"]).name: indexed.contents[
                    "SHA256"
                ]
            }
            if name in items:
                item, item_checksums = items[name]
                if index_checksums != item_checksums:
                    plan.replace.append(
                        PlanReplace[Packages](
                            name=name,
                            contents=indexed.contents,
                            component=indexed.component,
                            item=item,
                            ignore_unauthorized=indexed.ignore_unauthorized,
                        )
                    )
            else:
                plan.add.append(
                    PlanAdd[Packages](
                        name=name,
                        contents=indexed.contents,
                        component=indexed.component,
                        ignore_unauthorized=indexed.ignore_unauthorized,
                    )
                )

        for name in sorted(items.keys() - indexes.keys()):
            item, _ = items[name]
            plan.remove.append(item)

        return plan

    def add_binary(self, temp_path: Path, plan: PlanAdd[Packages]) -> None:
        """Download a binary package and add it to the collection."""
        with ExitStack() as stack:
            (temp_download := temp_path / "download").mkdir(parents=True)
            stack.callback(shutil.rmtree, temp_download)

            binary_artifact: Artifact | None = None
            if (
                self.collection_manager.parent_archive is not None
                and (
                    binary_from_archive
                    := self.collection_manager.parent_archive.manager.lookup(
                        f"binary-version:{plan.name}"
                    )
                )
                is not None
            ):
                # The parent archive already has this package, so we don't
                # need to download it again as long as its checksums match.
                assert binary_from_archive.artifact is not None
                if binary_from_archive.artifact.fileinartifact_set.filter(
                    path=PurePath(plan.contents["Filename"]).name,
                    file__size=int(plan.contents["Size"]),
                    file__sha256=binascii.unhexlify(plan.contents["SHA256"]),
                ).exists():
                    binary_artifact = binary_from_archive.artifact

            if binary_artifact is None:
                # We don't have this package anywhere usable, so we need to
                # download it again.
                if not self.fetch_archive_file(
                    temp_download,
                    plan.contents["Filename"],
                    expected_size=int(plan.contents["Size"]),
                    expected_sha256=plan.contents["SHA256"],
                    ignore_unauthorized=plan.ignore_unauthorized,
                ):
                    # Unauthorized, but ignored by filters.
                    return
                binary_package = BinaryPackage.create(
                    file=temp_download / plan.contents["Filename"]
                )
                binary_artifact = Artifact.objects.create_from_local_artifact(
                    binary_package,
                    self.workspace,
                    created_by_work_request=self.work_request,
                )
                srcpkg_name = binary_package.data.srcpkg_name
                srcpkg_version = binary_package.data.srcpkg_version
                source_item = self.collection_manager.lookup(
                    f"source-version:{srcpkg_name}_{srcpkg_version}"
                )
                # Add a built-using relationship to the corresponding source
                # if we can be confident of it.  Suites with
                # may_reuse_versions=True may have source packages replaced,
                # and in that case we can't be sure that the source package
                # version in the binary package's metadata is enough to
                # match it.
                if (
                    source_item is not None
                    and not self.collection_manager.may_reuse_versions
                ):
                    assert source_item.artifact is not None
                    ArtifactRelation.objects.create(
                        artifact=binary_artifact,
                        target=source_item.artifact,
                        type=ArtifactRelation.Relations.BUILT_USING,
                    )

            with self.collection_manager.allow_rewinding_versions():
                self.collection_manager.add_artifact(
                    binary_artifact,
                    user=self.work_request.created_by,
                    variables={
                        "component": plan.component,
                        "section": plan.contents["Section"],
                        "priority": plan.contents["Priority"],
                    },
                    replace=self.collection_manager.may_reuse_versions,
                )

    def update_binaries(self, temp_path: Path, plan: Plan[Packages]) -> None:
        """
        Update all binary packages in the collection.

        This may take a long time, so it commits transactions as it goes
        rather than taking a single long transaction.  If processing an item
        fails, then the results of earlier processing will remain visible.
        """
        for add in plan.add:
            with transaction.atomic():
                self.add_binary(temp_path, add)

        for replace in plan.replace:
            with transaction.atomic():
                self.collection_manager.remove_item(replace.item)
                self.add_binary(temp_path, replace)

        with transaction.atomic():
            for remove in plan.remove:
                self.collection_manager.remove_item(remove)

    @contextmanager
    def apt_config(self, temp_path: Path) -> Generator[None]:
        """Temporarily use a different APT configuration."""
        # We have to import this late, as otherwise mypy_django_plugin tries
        # to import it during early initialization and fails (since we can't
        # install it from PyPI and so it isn't present in mypy's virtual
        # environment).
        import apt_pkg

        try:
            os.environ["APT_CONFIG"] = self.make_apt_environment(temp_path)[
                "APT_CONFIG"
            ]
            # This looks like dictionary iteration, but
            # apt_pkg.Configuration isn't a real dictionary and isn't
            # iterable.
            for key in apt_pkg.config.keys():  # noqa: SIM118
                apt_pkg.config.clear(key)
            apt_pkg.init_config()
            yield
        finally:
            del os.environ["APT_CONFIG"]

    def plan_indexes(self, temp_path: Path) -> Plan[Deb822]:
        """Plan the update of all indexes in the collection."""
        # We don't mirror indexes for suites in the flat repository format,
        # because we don't currently have a reasonable way to serve them:
        # debusine.web.archives assumes a pooled layout, and we can't freely
        # convert between them because the Release file must agree with the
        # layout we serve.
        if self.data.suite.endswith("/"):
            return Plan[Deb822]()

        indexes: dict[str, Deb822] = {}
        for rel_path, (index_path, index_file) in self.get_fetched_index_files(
            temp_path
        ).items():
            indexes.setdefault(
                f"index:{rel_path}",
                Deb822(
                    {
                        "MetaKey": rel_path,
                        "Component": index_file.component,
                        "Filename": str(index_path),
                    }
                ),
            )

        for release_name in ("Release", "Release.gpg", "InRelease"):
            lists_path = self.get_apt_lists_path(temp_path)
            release_path = lists_path / self.get_apt_release_file_name(
                temp_path, release_name
            )
            if release_path.exists():
                indexes[f"index:{release_name}"] = Deb822(
                    {
                        "MetaKey": release_name,
                        "Component": "",
                        "Filename": str(release_path),
                    }
                )

        # Repository index files from the local collection
        items: dict[str, tuple[CollectionItem, dict[str, str]]] = {}
        for item in (
            self.collection.collection.child_items.active()
            .filter(
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.REPOSITORY_INDEX,
            )
            .only("name", "data", "artifact")
            .prefetch_related("artifact__fileinartifact_set__file")
        ):
            artifact = item.artifact
            assert artifact is not None
            items[item.name] = (
                item,
                {
                    file_in_artifact.path: file_in_artifact.file.sha256.hex()
                    for file_in_artifact in artifact.fileinartifact_set.all()
                },
            )

        plan = Plan[Deb822]()

        for name, paragraph in sorted(indexes.items()):
            rel_path = paragraph["MetaKey"]
            component = paragraph["Component"]
            index_path = Path(paragraph["Filename"])
            index_checksums = {
                PurePath(rel_path).name: (
                    calculate_hash(index_path, "sha256").hex()
                )
            }
            if name in items:
                item, item_checksums = items[name]
                if index_checksums != item_checksums:
                    plan.replace.append(
                        PlanReplace[Deb822](
                            name=rel_path,
                            contents=paragraph,
                            component=component,
                            item=item,
                        )
                    )
            else:
                plan.add.append(
                    PlanAdd[Deb822](
                        name=rel_path, contents=paragraph, component=component
                    )
                )

        for name in sorted(items.keys() - indexes.keys()):
            item, _ = items[name]
            plan.remove.append(item)

        return plan

    def add_index(self, *, name: str, paragraph: Deb822) -> None:
        """Add an index to the collection."""
        index = RepositoryIndex.create(
            file=Path(paragraph["Filename"]), path=name
        )
        index_artifact = Artifact.objects.create_from_local_artifact(
            index, self.workspace, created_by_work_request=self.work_request
        )
        self.collection_manager.add_artifact(
            index_artifact,
            user=self.work_request.created_by,
            variables={"path": name},
        )

    def update_indexes(self, plan: Plan[Deb822]) -> None:
        """
        Update all repository indexes in the collection.

        Indexes are normally relatively small compared to packages and have
        already been downloaded by apt, so it does this in a single
        transaction.
        """
        with transaction.atomic():
            for add in plan.add:
                self.add_index(name=add.name, paragraph=add.contents)

            for replace in plan.replace:
                self.collection_manager.remove_item(replace.item)
                self.add_index(name=replace.name, paragraph=replace.contents)

            for remove in plan.remove:
                self.collection_manager.remove_item(remove)

    @override
    def _execute(self) -> WorkRequestResults:
        """Execute the task."""
        with advisory_lock(
            (
                LockType.APT_MIRROR,
                # Only use the bottom 31 bits, in order that this fits into
                # PostgreSQL's int type.  In the unlikely event that we have
                # enough mirrored collections for there to be a collision,
                # then it just means that the colliding collections can't be
                # mirrored simultaneously.
                self.collection.collection_id & (2**31 - 1),
            ),
            wait=False,
        ) as acquired:
            if not acquired:
                raise LockError(
                    f"Another APTMirror task for {self.data.collection} is "
                    f"already running"
                )

            # This task may take a long time, so it commits transactions as
            # it goes rather than taking a single long transaction.  If
            # processing an item fails, then the results of earlier
            # processing will remain visible.
            with tempfile.TemporaryDirectory(
                prefix="debusine-aptmirror-"
            ) as temp_dir:
                temp_path = Path(temp_dir)
                self.fetch_meta_indexes(temp_path)
                self.fetch_indexes(temp_path)
                self.update_sources(temp_path, self.plan_sources(temp_path))
                self.update_binaries(temp_path, self.plan_binaries(temp_path))
                self.update_indexes(self.plan_indexes(temp_path))
            return WorkRequestResults.SUCCESS

    @override
    def get_label(self) -> str:
        """Return the task label."""
        return f"mirror {self.data.collection} from {self.data.url}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debian_pipeline workflow."""

from functools import partial
from typing import assert_never, override

from debusine.artifacts import SourcePackage
from debusine.artifacts.models import ArtifactCategory, DebianSourcePackage
from debusine.assets import KeyPurpose
from debusine.client.models import LookupChildType
from debusine.db.models import WorkRequest
from debusine.server.tasks.wait.models import ConfirmData
from debusine.server.workflows import (
    Workflow,
    WorkflowValidationError,
    workflow_utils,
)
from debusine.server.workflows.models import (
    DebianPipelineWorkflowData,
    DebianPipelineWorkflowDynamicData,
    MakeSignedSourceWorkflowData,
    PackagePublishWorkflowData,
    PackageUploadWorkflowData,
    QAFailOn,
    QAFailurePolicy,
    QAWorkflowData,
    RegressionTrackingUpdateQAResults,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.tasks.models import (
    BackendType,
    ExtraDebusineRepository,
    ExtraRepository,
    InputArtifact,
    InputArtifactSingle,
    LintianFailOnSeverity,
    LookupMultiple,
    LookupSingle,
    SbuildInput,
)


class DebianPipelineWorkflow(
    Workflow[DebianPipelineWorkflowData, DebianPipelineWorkflowDynamicData]
):
    """Debian pipeline workflow."""

    TASK_NAME = "debian_pipeline"

    @override
    def validate_input(self) -> None:
        """Thorough validation of input data."""
        if self.data.qa_suite is not None:
            try:
                self.work_request.lookup_single(
                    self.data.qa_suite, expect_type=LookupChildType.COLLECTION
                )
            except LookupError as e:
                raise WorkflowValidationError(str(e)) from e

    @override
    def compute_dynamic_data(self) -> DebianPipelineWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: package name of ``source_artifact``
        :source_artifact_id: id of the source artifact
        """
        source_artifact = workflow_utils.source_package(self)
        source_data = SourcePackage.create_data(source_artifact.data)

        return DebianPipelineWorkflowDynamicData(
            subject=source_data.name,
            parameter_summary=f"{source_data.name}_{source_data.version}",
            input_source_artifact_id=source_artifact.id,
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return workflow input artifacts (the source package)."""
        return [
            InputArtifactSingle(
                lookup=self.data.source_artifact,
                label="source_artifact",
                artifact_id=(
                    self.dynamic_data.input_source_artifact_id
                    if self.dynamic_data is not None
                    else None
                ),
            )
        ]

    def suite_has_active_indexes(self, suite_lookup: LookupSingle) -> bool:
        """Return True if this suite has an active ``InRelease`` file."""
        suite = self.work_request.lookup_single(
            suite_lookup, expect_type=LookupChildType.COLLECTION
        ).collection
        return (
            suite.child_items.active()
            .filter(
                category=ArtifactCategory.REPOSITORY_INDEX,
                data__path="InRelease",
            )
            .exists()
        )

    def get_extra_repositories(self) -> list[ExtraRepository] | None:
        """Decide which extra repositories to use for this workflow."""
        extra_repositories = (
            None
            if self.data.extra_repositories is None
            else list(self.data.extra_repositories)
        )
        if self.data.suite is not None and self.suite_has_active_indexes(
            self.data.suite
        ):
            if extra_repositories is None:
                extra_repositories = []
            extra_repositories.insert(
                0, ExtraDebusineRepository(suite=self.data.suite)
            )
        return extra_repositories

    @override
    def populate(self) -> None:  # noqa: C901
        """Create work requests."""
        assert self.dynamic_data is not None
        assert self.dynamic_data.subject is not None

        source_data = workflow_utils.source_package_data(self)

        if (data_archs := self.data.architectures) is not None:
            architectures = set(data_archs)
        else:
            architectures = workflow_utils.get_available_architectures(
                self, vendor=self.data.vendor, codename=self.data.codename
            )

        if (
            data_archs_allowlist := self.data.architectures_allowlist
        ) is not None:
            architectures.intersection_update(data_archs_allowlist)

        if (
            data_archs_denylist := self.data.architectures_denylist
        ) is not None:
            architectures.difference_update(data_archs_denylist)

        target_distribution = f"{self.data.vendor}:{self.data.codename}"

        effective_architectures = sorted(architectures)

        extra_repositories = self.get_extra_repositories()

        self._populate_sbuild(
            source_artifact=self.data.source_artifact,
            target_distribution=target_distribution,
            backend=self.data.sbuild_backend,
            architectures=effective_architectures,
            environment_variant=self.data.sbuild_environment_variant,
            extra_repositories=extra_repositories,
            signing_template_names=self.data.signing_template_names,
        )

        # sbuild workflow promises the build-{arch} artifacts
        binary_artifacts = LookupMultiple.model_validate(
            {
                "collection": "internal@collections",
                "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                "name__startswith": "build-",
            }
        )

        to_upload = [
            (
                self.data.source_artifact,
                binary_artifacts,
                self.dynamic_data.subject,
                False,
            )
        ]

        # Work request that post-QA steps in the pipeline should depend on
        # for QA confirmation.  Depending on the task data, this may be
        # automatic or manual.
        qa_confirmation: WorkRequest | None = None

        qa: WorkRequest | None = None
        if (
            self.data.enable_autopkgtest
            or self.data.enable_reverse_dependencies_autopkgtest
            or self.data.enable_lintian
            or self.data.enable_piuparts
            or self.data.enable_debdiff
            or self.data.enable_blhc
        ):
            populate_qa = partial(
                self._populate_qa,
                source_artifact=self.data.source_artifact,
                vendor=self.data.vendor,
                codename=self.data.codename,
                architectures=effective_architectures,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                extra_repositories=extra_repositories,
                qa_suite=self.data.qa_suite,
                enable_autopkgtest=self.data.enable_autopkgtest,
                autopkgtest_backend=self.data.autopkgtest_backend,
                enable_reverse_dependencies_autopkgtest=(
                    self.data.enable_reverse_dependencies_autopkgtest
                ),
                enable_lintian=self.data.enable_lintian,
                lintian_backend=self.data.lintian_backend,
                lintian_fail_on_severity=self.data.lintian_fail_on_severity,
                enable_piuparts=self.data.enable_piuparts,
                piuparts_backend=self.data.piuparts_backend,
                piuparts_environment=self.data.piuparts_environment,
                enable_blhc=self.data.enable_blhc,
            )

            if self.data.enable_regression_tracking:
                reference_source_lookup = (
                    f"{self.data.qa_suite}/source:{source_data.name}"
                )
                try:
                    reference_source = self.work_request.lookup_single(
                        reference_source_lookup,
                        expect_type=LookupChildType.ARTIFACT,
                    ).artifact
                except KeyError:
                    # This source package does not exist in qa_suite.  Skip
                    # regression tracking.
                    pass
                else:
                    reference_source_data = reference_source.create_data()
                    assert isinstance(
                        reference_source_data, DebianSourcePackage
                    )
                    reference_name = reference_source_data.name
                    reference_version = reference_source_data.version
                    populate_qa(
                        prefix="reference-qa-result|",
                        source_artifact=reference_source.id,
                        binary_artifacts=LookupMultiple.model_validate(
                            {
                                "collection": self.data.qa_suite,
                                "category": ArtifactCategory.BINARY_PACKAGE,
                                "data__srcpkg_name": reference_name,
                                "data__srcpkg_version": reference_version,
                            }
                        ),
                        enable_regression_tracking=False,
                        reference_qa_results=(
                            self.data.regression_tracking_qa_results
                        ),
                        update_qa_results=RegressionTrackingUpdateQAResults.YES,
                        # The debdiff workflow doesn't contribute any QA
                        # results, so don't waste time running it here.
                        enable_debdiff=False,
                        fail_on=QAFailOn.NEVER,
                        workflow_data=WorkRequestWorkflowData(
                            display_name=(
                                f"QA (reference tests for "
                                f"{reference_name}_{reference_version})"
                            ),
                            step="qa-reference",
                            expanded=False,
                        ),
                    )
                    qa = populate_qa(
                        source_artifact=self.data.source_artifact,
                        binary_artifacts=binary_artifacts,
                        reference_source_artifact=reference_source.id,
                        enable_regression_tracking=True,
                        reference_qa_results=(
                            self.data.regression_tracking_qa_results
                        ),
                        reference_prefix="reference-qa-result|",
                        fail_on=QAFailOn.REGRESSION,
                        enable_debdiff=self.data.enable_debdiff,
                    )

            if qa is None:
                qa = populate_qa(
                    source_artifact=self.data.source_artifact,
                    binary_artifacts=binary_artifacts,
                    enable_debdiff=self.data.enable_debdiff,
                )

            match self.data.qa_failure_policy:
                case QAFailurePolicy.IGNORE:
                    pass
                case QAFailurePolicy.FAIL:
                    qa_confirmation = qa
                case QAFailurePolicy.CONFIRM:
                    qa_confirmation = self._populate_confirmation(
                        auto_confirm_if_no_failure=True,
                        confirm_label=(
                            "Continue the workflow despite QA failures"
                        ),
                    )
                    qa_confirmation.add_dependency(qa)
                case _ as unreachable:
                    assert_never(unreachable)

        if qa_confirmation is None and self.data.enable_confirmation:
            qa_confirmation = self._populate_confirmation()

        if (
            self.data.enable_make_signed_source
            and self.data.signing_template_names is not None
        ):
            if self.data.make_signed_source_purpose is None:
                raise WorkflowValidationError(
                    '"make_signed_source_purpose" must be set when '
                    'signing the source'
                )

            if self.data.make_signed_source_key is None:
                raise WorkflowValidationError(
                    '"make_signed_source_key" must be set when '
                    'signing the source'
                )

            for (
                upload_source_artifact,
                upload_binary_artifacts,
                upload_subject,
            ) in self._populate_make_signed_source(
                binary_artifacts=binary_artifacts,
                vendor=self.data.vendor,
                codename=self.data.codename,
                architectures=effective_architectures,
                purpose=self.data.make_signed_source_purpose,
                key=self.data.make_signed_source_key,
                sbuild_backend=self.data.sbuild_backend,
                signing_template_names=self.data.signing_template_names,
                qa_confirmation=qa_confirmation,
            ):
                to_upload.append(
                    (
                        upload_source_artifact,
                        upload_binary_artifacts,
                        upload_subject,
                        True,
                    )
                )

        if self.data.enable_upload:
            for (
                upload_source_artifact,
                upload_binary_artifacts,
                upload_subject,
                signed_source,
            ) in to_upload:
                self._populate_package_upload(
                    source_artifact=upload_source_artifact,
                    binary_artifacts=upload_binary_artifacts,
                    include_source=self.data.upload_include_source,
                    include_binaries=self.data.upload_include_binaries,
                    signed_source=signed_source,
                    merge_uploads=self.data.upload_merge_uploads,
                    since_version=self.data.upload_since_version,
                    key=self.data.upload_key,
                    binary_key=self.data.upload_binary_key,
                    require_signature=self.data.upload_require_signature,
                    target=self.data.upload_target,
                    delayed_days=self.data.upload_delayed_days,
                    vendor=self.data.vendor,
                    codename=self.data.codename,
                    subject=upload_subject,
                    qa_confirmation=qa_confirmation,
                )

        if self.data.suite:
            for (
                upload_source_artifact,
                upload_binary_artifacts,
                upload_subject,
                _,
            ) in to_upload:
                self._populate_package_publish(
                    source_artifact=upload_source_artifact,
                    binary_artifacts=upload_binary_artifacts,
                    target_suite=self.data.suite,
                    replace=self.data.publish_replace,
                    subject=upload_subject,
                    qa_confirmation=qa_confirmation,
                )

    def _populate_sbuild(
        self,
        *,
        source_artifact: LookupSingle,
        target_distribution: str,
        backend: BackendType,
        architectures: list[str],
        environment_variant: str | None,
        extra_repositories: list[ExtraRepository] | None = None,
        signing_template_names: dict[str, list[str]],
    ) -> None:
        """Create work request for sbuild workflow."""
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="sbuild",
            task_data=SbuildWorkflowData(
                input=SbuildInput(source_artifact=source_artifact),
                target_distribution=target_distribution,
                backend=backend,
                architectures=architectures,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                environment_variant=environment_variant,
                extra_repositories=extra_repositories,
                signing_template_names=signing_template_names,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="sbuild",
                step="sbuild",
            ),
        )
        self.orchestrate_child(wr)

    def _populate_qa(
        self,
        *,
        prefix: str = "",
        reference_prefix: str = "",
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        reference_source_artifact: LookupSingle | None = None,
        vendor: str,
        codename: str,
        architectures: list[str],
        arch_all_build_architecture: str,
        extra_repositories: list[ExtraRepository] | None = None,
        qa_suite: LookupSingle | None,
        reference_qa_results: LookupSingle | None = None,
        enable_regression_tracking: bool = False,
        update_qa_results: RegressionTrackingUpdateQAResults = (
            RegressionTrackingUpdateQAResults.NO
        ),
        enable_autopkgtest: bool,
        autopkgtest_backend: BackendType,
        enable_reverse_dependencies_autopkgtest: bool,
        enable_lintian: bool,
        lintian_backend: BackendType,
        lintian_fail_on_severity: LintianFailOnSeverity,
        enable_piuparts: bool,
        piuparts_backend: BackendType,
        piuparts_environment: LookupSingle | None,
        enable_debdiff: bool,
        enable_blhc: bool,
        fail_on: QAFailOn = QAFailOn.FAILURE,
        workflow_data: WorkRequestWorkflowData | None = None,
    ) -> WorkRequest:
        """Create work request for qa workflow."""
        data = QAWorkflowData(
            prefix=prefix,
            reference_prefix=reference_prefix,
            source_artifact=source_artifact,
            binary_artifacts=binary_artifacts,
            reference_source_artifact=reference_source_artifact,
            vendor=vendor,
            codename=codename,
            architectures=architectures,
            arch_all_build_architecture=arch_all_build_architecture,
            extra_repositories=extra_repositories,
            qa_suite=qa_suite,
            reference_qa_results=reference_qa_results,
            enable_regression_tracking=enable_regression_tracking,
            update_qa_results=update_qa_results,
            enable_autopkgtest=enable_autopkgtest,
            autopkgtest_backend=autopkgtest_backend,
            enable_reverse_dependencies_autopkgtest=(
                enable_reverse_dependencies_autopkgtest
            ),
            enable_lintian=enable_lintian,
            lintian_backend=lintian_backend,
            lintian_fail_on_severity=lintian_fail_on_severity,
            enable_piuparts=enable_piuparts,
            piuparts_backend=piuparts_backend,
            enable_debdiff=enable_debdiff,
            enable_blhc=enable_blhc,
            fail_on=fail_on,
        )
        if piuparts_environment:
            data.piuparts_environment = piuparts_environment
        if workflow_data is None:
            workflow_data = WorkRequestWorkflowData(
                display_name="QA", step="qa"
            )
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="qa", task_data=data, workflow_data=workflow_data
        )
        # The qa workflow's children will have dependencies on the work
        # requests creating binary_artifacts, but the qa workflow itself
        # doesn't need that in order to populate itself.
        self.orchestrate_child(wr)
        return wr

    def _populate_confirmation(
        self,
        *,
        auto_confirm_if_no_failure: bool = False,
        confirm_label: str = "Continue the workflow",
    ) -> WorkRequest:
        """Create a work request acting as a confirmation step."""
        wr, _ = self.work_request_ensure_child_wait(
            task_name="confirm",
            task_data=ConfirmData(
                auto_confirm_if_no_failure=auto_confirm_if_no_failure,
                confirm_label=confirm_label,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="Confirm", step="confirm", needs_input=True
            ),
        )
        return wr

    def _populate_make_signed_source(
        self,
        *,
        vendor: str,
        codename: str,
        binary_artifacts: LookupMultiple,
        architectures: list[str],
        purpose: KeyPurpose,
        key: str,
        sbuild_backend: BackendType,
        signing_template_names: dict[str, list[str]],
        qa_confirmation: WorkRequest | None = None,
    ) -> list[tuple[LookupSingle, LookupMultiple, str]]:
        """Create work request for make signed source workflow."""
        signing_artifacts: list[LookupSingle] = []
        to_upload: list[tuple[LookupSingle, LookupMultiple, str]] = []
        for arch, names in sorted(signing_template_names.items()):
            for name in names:
                # The promises for the signing_artifacts are created
                # by Sbuild workflow
                signing_artifact = (
                    f"internal@collections/name:signing-template-{arch}-{name}"
                )
                self.work_request.lookup_single(
                    signing_artifact,
                    expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
                )

                signing_artifacts.append(signing_artifact)
                to_upload.append(
                    (
                        f"internal@collections/name:"
                        f"signed-source-{arch}-{name}",
                        LookupMultiple.model_validate(
                            {
                                "collection": "internal@collections",
                                "child_type": (
                                    LookupChildType.ARTIFACT_OR_PROMISE
                                ),
                                "name__startswith": (
                                    f"signed-source-{arch}-{name}|build-"
                                ),
                            }
                        ),
                        f"{arch}/{name}",
                    )
                )

        signing_template_artifacts = LookupMultiple.model_validate(
            signing_artifacts
        )
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="make_signed_source",
            task_data=MakeSignedSourceWorkflowData(
                binary_artifacts=binary_artifacts,
                signing_template_artifacts=signing_template_artifacts,
                vendor=vendor,
                codename=codename,
                architectures=architectures,
                purpose=purpose,
                key=key,
                sbuild_backend=sbuild_backend,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="make signed source", step="make_signed_source"
            ),
        )

        # If confirmation is enabled, then the make_signed_source workflow
        # will not be populated until confirmation (to avoid having to poke
        # around in its internals to find the first work request for each
        # architecture).
        if qa_confirmation is not None:
            wr.add_dependency(qa_confirmation)

        # The make_signed_source workflow's children will have dependencies
        # on the work requests creating binary_artifacts and
        # signing_artifacts, but the make_signed_source workflow itself
        # doesn't need that in order to populate itself.
        self.orchestrate_child(wr)

        return to_upload

    def _populate_package_upload(
        self,
        *,
        binary_artifacts: LookupMultiple,
        source_artifact: LookupSingle,
        include_source: bool,
        include_binaries: bool,
        signed_source: bool,
        merge_uploads: bool,
        since_version: str | None,
        key: str | None,
        binary_key: str | None,
        require_signature: bool,
        target: str,
        delayed_days: int | None,
        vendor: str,
        codename: str,
        subject: str,
        qa_confirmation: WorkRequest | None = None,
    ) -> None:
        """Create work request for package upload workflow."""
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="package_upload",
            task_data=PackageUploadWorkflowData(
                source_artifact=(
                    source_artifact if include_source or signed_source else None
                ),
                binary_artifacts=(
                    binary_artifacts if include_binaries else LookupMultiple(())
                ),
                merge_uploads=merge_uploads,
                since_version=since_version,
                target_distribution=self.data.upload_target_distribution,
                key=key,
                binary_key=binary_key,
                require_signature=require_signature,
                target=target,
                delayed_days=delayed_days,
                vendor=vendor,
                codename=codename,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"package upload ({subject})",
                step=f"package_upload/{subject}",
            ),
        )

        # If confirmation is enabled, then the package_upload workflow will
        # not be populated until confirmation (to avoid having to poke
        # around in its internals to find the first work request).
        if qa_confirmation is not None:
            wr.add_dependency(qa_confirmation)

        # The package_upload workflow's children will have dependencies on
        # the work requests creating source_artifact and binary_artifacts,
        # but the package_upload workflow itself doesn't normally need that
        # in order to populate itself.  An exception is if we're uploading
        # signed source with binaries, in which case the make_signed_source
        # workflow will only be able to create even the necessary promises
        # once the signed source package has been assembled.
        if signed_source and include_binaries and binary_artifacts:
            self.requires_artifact(wr, source_artifact)

        self.orchestrate_child(wr)

    def _populate_package_publish(
        self,
        *,
        binary_artifacts: LookupMultiple,
        source_artifact: LookupSingle,
        target_suite: LookupSingle,
        replace: bool,
        subject: str,
        qa_confirmation: WorkRequest | None = None,
    ) -> None:
        """Create work request for package publish workflow."""
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="package_publish",
            task_data=PackagePublishWorkflowData(
                source_artifact=source_artifact,
                binary_artifacts=binary_artifacts,
                target_suite=target_suite,
                replace=replace,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"publish packages ({subject})",
                step=f"package_publish/{subject}",
            ),
        )

        # If confirmation is enabled, then the package_publish workflow will
        # not be populated until confirmation (to avoid having to poke
        # around in its internals to find the first work request).
        if qa_confirmation is not None:
            wr.add_dependency(qa_confirmation)

        # The package_publish workflow needs to split up debian:upload
        # binary artifacts into individual debian:binary-package artifacts
        # when preparing data for the CopyCollectionItems task, so its input
        # artifacts must exist before it can populate itself.
        self.requires_artifact(wr, source_artifact)
        self.requires_artifact(wr, binary_artifacts)

        self.orchestrate_child(wr)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine command line interface, archive management commands.

These are "porcelain": they implement opinionated operations on
:collection:`debian:archive` collections and their contents, rather than
basic "plumbing" operations.
"""

# Commands need to be imported in order to be available to cli.py.
# They are registered via Command.__init_subclass__.

import debusine.client.commands.archives.suites  # noqa: F401
import debusine.client.commands.archives.tokens  # noqa: F401



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-admin command to create workflows.

This is similar to the API endpoint, but it doesn't require a token so it's
easier to use in simple automation.
"""

import argparse
import sys
from typing import Any, NoReturn, override

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management import CommandError, CommandParser
from django.db import transaction

from debusine.db.context import context
from debusine.db.models import (
    SYSTEM_USER_NAME,
    User,
    WorkRequest,
    WorkflowTemplate,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import get_workspace


class Command(DebusineBaseCommand):
    """Command to create a workflow."""

    help = "Create a new workflow"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument("template_name", help="Template name")
        parser.add_argument(
            "--created-by",
            help="Name of the user creating the workflow",
            default=SYSTEM_USER_NAME,
        )
        parser.add_argument(
            "--workspace",
            metavar="scope/name",
            help="Workspace",
            default=settings.DEBUSINE_DEFAULT_WORKSPACE,
        )
        parser.add_argument(
            "--data",
            type=argparse.FileType("r"),
            help=(
                "File path (or - for stdin) to read the data for the work "
                "request. YAML format. Defaults to stdin."
            ),
            default="-",
        )

    @override
    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        if options["data"] != sys.stdin:
            options["data"].close()

    @override
    @context.local()
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Create the workflow."""
        template_name = options["template_name"]
        created_by_name = options["created_by"]
        data = self.parse_yaml_data(options["data"].read()) or {}

        workspace = get_workspace(options["workspace"])

        try:
            created_by = get_user_model().objects.get(username=created_by_name)
        except User.DoesNotExist:
            raise CommandError(
                f'User "{created_by_name}" not found', returncode=3
            )

        try:
            template = WorkflowTemplate.objects.get(
                name=template_name, workspace=workspace
            )
        except WorkflowTemplate.DoesNotExist:
            raise CommandError(
                f'Workflow template "{template_name}" not found', returncode=3
            )

        context.set_permission_context_scoped(
            workspace.scope, PermissionContext.for_ui_user(created_by)
        )
        workspace.set_current()

        with context.disable_permission_checks(), transaction.atomic():
            try:
                workflow = WorkRequest.objects.create_workflow(
                    template=template, data=data
                )
                workflow.mark_pending()
            except Exception as e:
                raise CommandError(
                    f"Failed to create workflow: {e}", returncode=3
                )

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for model objects."""

# Import modules so their helpers can register themselves
from debusine.web.views.ui import (
    artifacts,  # noqa: F401
    auth,  # noqa: F401
    collection_relations,  # noqa: F401
    collections,  # noqa: F401
    scopes,  # noqa: F401
    work_requests,  # noqa: F401
    worker_pools,  # noqa: F401
    workers,  # noqa: F401
    workspaces,  # noqa: F401
)
from debusine.web.views.ui.base import UI

__all__ = ["UI"]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Add the runtime_parameters field to WorkflowTemplate."""

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0032_rename_task_data_workflowtemplate_static_parameters'),
    ]

    operations = [
        migrations.AddField(
            model_name='workflowtemplate',
            name='runtime_parameters',
            field=models.JSONField(
                blank=True,
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""``debian_pipeline``: Rename ``publish_target`` to ``suite``."""

from django.db import migrations

from debusine.artifacts.models import TaskTypes
from debusine.db.migrations._utils import (
    make_work_request_task_data_field_renamer,
    make_workflow_template_task_data_field_renamer,
)


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0025_base_manager_name"),
    ]

    operations = [
        make_work_request_task_data_field_renamer(
            TaskTypes.WORKFLOW, "debian_pipeline", [("publish_target", "suite")]
        ),
        make_workflow_template_task_data_field_renamer(
            "debian_pipeline", [("publish_target", "suite")]
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Server-side Debusine tasks.

The tasks in this package run in a Celery worker.
"""

from debusine.server.tasks.base import BaseServerTask  # isort: split

# Sub-tasks need to be imported in order to be available to DBTask
# (e.g. for DBTask.is_valid_task_name). They are registered via
# DBTask.__init_subclass__.
from debusine.server.tasks.aptmirror import APTMirror
from debusine.server.tasks.base import (
    ServerTaskPermissionDenied,
    analyze_celery_worker_tasks,
    get_provided_celery_worker_tags,
)
from debusine.server.tasks.copy_collection_items import CopyCollectionItems
from debusine.server.tasks.create_child_workspace import CreateChildWorkspace
from debusine.server.tasks.generate_suite_indexes import GenerateSuiteIndexes
from debusine.server.tasks.noop import ServerNoop
from debusine.server.tasks.package_upload import PackageUpload

__all__ = [
    "APTMirror",
    "BaseServerTask",
    "CopyCollectionItems",
    "CreateChildWorkspace",
    "GenerateSuiteIndexes",
    "PackageUpload",
    "ServerNoop",
    "ServerTaskPermissionDenied",
    "analyze_celery_worker_tasks",
    "get_provided_celery_worker_tags",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the debdiff task support on the worker."""

import itertools
import re
from typing import override
from unittest.mock import call

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts import DebDiffArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
    WorkRequestResults,
)
from debusine.client.models import LookupSingleResponseArtifact
from debusine.tasks import DebDiff, TaskConfigError
from debusine.tasks.models import (
    BaseTaskInputArtifact,
    DebDiffDynamicData,
    InputArtifactSingle,
)
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_remote_artifact,
    create_system_tarball_data,
)


class DebDiffTests(ExternalTaskHelperMixin[DebDiff], TestCase):
    """Test the DebDiff class."""

    SAMPLE_TASK_DATA = {
        "input": {"source_artifacts": [421, 123]},
        "build_architecture": "amd64",
        "environment": "debian/match:codename=bookworm",
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact_list(
            "source_artifacts",
            InputArtifactSingle(
                lookup=421,
                label="input.source_artifacts",
                artifact_id=421,
                category=ArtifactCategory.SOURCE_PACKAGE,
                data=DebianSourcePackage(
                    name="hello", version="1.0", type="dpkg", dsc_fields={}
                ),
            ),
            InputArtifactSingle(
                lookup=123,
                label="input.source_artifacts",
                artifact_id=123,
                category=ArtifactCategory.SOURCE_PACKAGE,
                data=DebianSourcePackage(
                    name="hello", version="1.1", type="dpkg", dsc_fields={}
                ),
            ),
        )
        self.resolver.set_field_artifact_difflist("binary_artifacts")

    def test_configure_fails_with_missing_required_data(
        self,
    ) -> None:
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"source_artifacts": {}})

    def test_configure_with_extra_flag(self) -> None:
        """Configuration included "extra_flags". No exception raised."""
        extra_flags = ["--dirs", "--nocontrol"]

        self.configure_task(override={"extra_flags": extra_flags})

        self.assertEqual(self.task.data.extra_flags, extra_flags)

    def test_configure_with_invalid_flag(self) -> None:
        """Invalid extra flags are rejected."""
        extra_flags = ["--unknown-flag"]

        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"extra_flags": extra_flags})

    def test_configure_additional_properties_in_input(self) -> None:
        """Assert no additional properties in task_data input."""
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(TaskConfigError, error_msg):
            self.configure_task(
                task_data={
                    "input": {
                        "artifact": 5,
                        "some_additional_property": 87,
                    }
                }
            )

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:debdiff:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.SOURCE_PACKAGE_PREFIX + "hello"},
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
            },
        )

    def test_compute_scheduler_tags_no_source(self) -> None:
        self.resolver.set_field_artifact_list("source_artifacts")
        self.resolver.set_field_artifact_difflist(
            "binary_artifacts",
            [
                BaseTaskInputArtifact(
                    artifact_id=500,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Version": "1.0-1",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                )
            ],
            [
                BaseTaskInputArtifact(
                    artifact_id=501,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Version": "1.1-1",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                )
            ],
        )

        provided, required = self.compute_scheduler_tags()
        self.assertEqual(provided.sets, {})
        self.assertEqual(
            required.sets,
            {
                wtags.ProvenanceRequired.SYSTEM: {
                    wtags.BUILD_ARCH_PREFIX + "amd64",
                    wtags.EXECUTOR_PREFIX + "unshare",
                }
            },
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebDiffDynamicData(
                environment_id=1,
                input_source_artifacts_ids=[421, 123],
                input_binary_artifacts_ids=None,
                subject="source:hello",
                parameter_summary="hello: 1.0 → 1.1 (source)",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "source:hello")

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_compute_dynamic_subject_binary_artifacts(self) -> None:
        self.resolver.set_field_artifact_list("source_artifacts")
        self.resolver.set_field_artifact_difflist(
            "binary_artifacts",
            [
                BaseTaskInputArtifact(
                    artifact_id=500,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Version": "1.0-1",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
            ],
            [
                BaseTaskInputArtifact(
                    artifact_id=501,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Version": "1.1-1",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
            ],
        )
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebDiffDynamicData(
                environment_id=1,
                input_source_artifacts_ids=None,
                input_binary_artifacts_ids=[[500], [501]],
                subject="binary:hello",
                parameter_summary="hello: 1.0-1 → 1.1-1 (amd64)",
            ),
        )

    def test_compute_dynamic_subject_is_none(self) -> None:
        """
        Test compute_dynamic_data set subject is None.

        There is no source package, and binary packages are from different
        source packages.
        """
        self.resolver.set_field_artifact_list("source_artifacts")
        self.resolver.set_field_artifact_difflist(
            "binary_artifacts",
            [
                BaseTaskInputArtifact(
                    artifact_id=500,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Version": "1.0-1",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
            ],
            [
                BaseTaskInputArtifact(
                    artifact_id=501,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "linux-base",
                            "Version": "1.1-1",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
            ],
        )

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebDiffDynamicData(
                environment_id=1,
                input_source_artifacts_ids=None,
                input_binary_artifacts_ids=[[500], [501]],
                subject=None,
            ),
        )

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            DebDiff.inputs["source_artifacts"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.SOURCE_PACKAGE,),
        )
        self.assertEqual(
            DebDiff.inputs["binary_artifacts"].categories,  # type: ignore[attr-defined]
            (
                ArtifactCategory.UPLOAD,
                ArtifactCategory.BINARY_PACKAGE,
            ),
        )

    def test_compute_dynamic_data_raise_task_config_error_categories(
        self,
    ) -> None:
        """Raise TaskConfigError: invalid categories binary artifacts."""
        self.resolver.set_field_artifact_list("source_artifacts")
        self.resolver.set_field_artifact_difflist(
            "binary_artifacts",
            [
                BaseTaskInputArtifact(
                    artifact_id=500,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
            ],
            [
                BaseTaskInputArtifact(
                    artifact_id=501,
                    category=ArtifactCategory.BINARY_PACKAGE,
                    data=DebianBinaryPackage(
                        srcpkg_name="linux-base",
                        srcpkg_version="1.0",
                        deb_fields={},
                        deb_control_files=[],
                    ),
                ),
            ],
        )

        with self.assertRaisesRegex(
            TaskConfigError,
            r'^All binary artifacts must have the same category. '
            r'Found "debian:upload" and "debian:binary-package"$',
        ):
            self.compute_dynamic_data(self.task)

    def test_compute_dynamic_data_raise_invalid_category(self) -> None:
        """Raise TaskConfigError: invalid categories binary artifacts."""
        self.resolver.set_field_artifact_list("source_artifacts")
        self.resolver.set_field_artifact_difflist(
            "binary_artifacts",
            [
                BaseTaskInputArtifact(
                    artifact_id=500,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
                BaseTaskInputArtifact(
                    artifact_id=501,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
            ],
            [
                BaseTaskInputArtifact(
                    artifact_id=502,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
                BaseTaskInputArtifact(
                    artifact_id=503,
                    category=ArtifactCategory.UPLOAD,
                    data=DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "amd64",
                            "Source": "hello",
                            "Files": [{"name": "hello.deb"}],
                        },
                    ),
                ),
            ],
        )

        with self.assertRaisesRegex(
            TaskConfigError,
            r"^If binary_artifacts source and new contain more "
            r"than one artifact all must be of category debian:binary-package. "
            r"Found: debian:upload$",
        ):
            self.compute_dynamic_data(self.task)

    def test_compute_dynamic_data_raise_binary_cannot_be_zero(self) -> None:
        self.resolver.set_field_artifact_list("source_artifacts")
        self.resolver.set_field_artifact_difflist("binary_artifacts", [], [])

        with self.assertRaisesRegex(
            TaskConfigError,
            re.escape(
                "input.binary_artifacts[0] and input.binary_artifacts[1] "
                "cannot have zero artifacts (got 0 and 0)"
            ),
        ):
            self.compute_dynamic_data(self.task)

    def test_get_input_artifacts_ids_dynamic_dynamic_data_none(self) -> None:
        self.task.dynamic_data = None
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

    def test_get_input_artifacts_ids_source_artifacts(self) -> None:
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_source_artifacts_ids=[1, 2],
            input_binary_artifacts_ids=None,
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2])

    def test_get_input_artifacts_ids_binary_artifacts(self) -> None:
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_source_artifacts_ids=None,
            input_binary_artifacts_ids=[[1, 2], [3, 4]],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2, 3, 4])

    def test_fetch_input_source(self) -> None:
        """Test fetching different source inputs."""
        self.configure_task(override={"input": {'source_artifacts': [1, 2]}})
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_source_artifacts_ids=[1, 2],
            input_binary_artifacts_ids=None,
        )
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponseArtifact(
            artifact=1
        )
        debusine_mock.download_artifact.return_value = True
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=2, workspace=self.task.workspace_name
        )

        self.assertTrue(self.task.fetch_input(download_directory))

    def test_fetch_input_binary(self) -> None:
        """Test fetching different binary inputs."""
        self.configure_task(
            override={"input": {'binary_artifacts': [[1, 2], [3, 4]]}}
        )
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_source_artifacts_ids=None,
            input_binary_artifacts_ids=[[1, 2], [3, 4]],
        )
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponseArtifact(
            artifact=1
        )
        debusine_mock.download_artifact.return_value = True
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=2, workspace=self.task.workspace_name
        )

        self.assertTrue(self.task.fetch_input(download_directory))

    def test_configure_for_execution_binaries(self) -> None:
        """configure_for_execution() with .changes files: return True."""
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_source_artifacts_ids=None,
            input_binary_artifacts_ids=[[1], [3]],
        )
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()

        (artifact1_dir := download_directory / "artifact_1").mkdir()
        (artifact1_dir / "file1.changes").write_text("")

        (artifact2_dir := download_directory / "artifact_2").mkdir()
        (artifact2_dir / "file2.changes").write_text("")

        self.task._original_dirs = [artifact1_dir]
        self.task._new_targets_dirs = [artifact2_dir]

        self.assertTrue(self.task.configure_for_execution(download_directory))

        self.assertEqual(
            self.task._original_targets, [artifact1_dir / "file1.changes"]
        )
        self.assertEqual(
            self.task._new_targets, [artifact2_dir / "file2.changes"]
        )

    def test_configure_for_execution_sources(self) -> None:
        """configure_for_execution() with .dsc files: return True."""
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_source_artifacts_ids=[1, 2],
        )
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()

        (artifact1_dir := download_directory / "artifact_1").mkdir()
        (artifact1_dir / "file1.dsc").write_text("")

        (artifact2_dir := download_directory / "artifact_2").mkdir()
        (artifact2_dir / "file2.dsc").write_text("")

        self.task._original_dirs = [artifact1_dir]
        self.task._new_targets_dirs = [artifact2_dir]

        self.assertTrue(self.task.configure_for_execution(download_directory))

        self.assertEqual(
            self.task._original_targets, [artifact1_dir / "file1.dsc"]
        )
        self.assertEqual(self.task._new_targets, [artifact2_dir / "file2.dsc"])

    def test_cmdline_with_extra_flags(self) -> None:
        """Cmdline add extra_flags."""
        extra_flags = ["--dirs", "--nocontrol"]
        self.configure_task(override={"extra_flags": extra_flags})
        original = self.create_temporary_file(suffix=".dsc")
        new = self.create_temporary_file(suffix=".dsc")

        self.task._original_targets = [original]
        self.task._new_targets = [new]

        cmdline = self.task._cmdline()

        self.assertEqual(
            cmdline,
            ["debdiff", "--dirs", "--nocontrol", str(original), str(new)],
        )

    def test_cmdline_with_from_to(self) -> None:
        """Cmdline use "from" and "to"."""
        self.task._original_targets = [
            self.create_temporary_file(suffix=".deb"),
            self.create_temporary_file(suffix=".deb"),
        ]
        self.task._new_targets = [
            self.create_temporary_file(suffix=".deb"),
            self.create_temporary_file(suffix=".udeb"),
        ]

        cmdline = self.task._cmdline()

        self.assertEqual(
            cmdline,
            [
                "debdiff",
                "--from",
                *[str(path) for path in self.task._original_targets],
                "--to",
                *[str(path) for path in self.task._new_targets],
            ],
        )

    def test_task_result_empty_file_success(self) -> None:
        """task_result() for an empty file return SUCCESS."""
        directory = self.create_temporary_directory()
        (directory / DebDiff.CAPTURE_OUTPUT_FILENAME).write_text("")

        self.configure_task()
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_failure_exit_code(self) -> None:
        """task_result() for an exit code of 255 return FAILURE."""
        directory = self.create_temporary_directory()
        (directory / DebDiff.CAPTURE_OUTPUT_FILENAME).write_text("")

        self.configure_task()
        self.assertEqual(
            self.task.task_result(255, directory), WorkRequestResults.FAILURE
        )

    def test_upload_artifacts(self) -> None:
        """upload_artifact() and relation_create() is called."""
        exec_dir = self.create_temporary_directory()

        # Create file that will be attached when uploading the artifacts
        debdiff_output = exec_dir / self.task.CAPTURE_OUTPUT_FILENAME
        debdiff_output.write_text("1c1\n< Source: foo\n---\n> Source: bar\n")

        # Used to verify the relations
        self.task._source_artifacts_ids = [1]
        binary_artifact_id_1 = 10
        binary_artifact_id_2 = 11
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_binary_artifacts_ids=[
                [binary_artifact_id_1],
                [binary_artifact_id_2],
            ],
            parameter_summary="A → B",
        )

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        workspace_name = "testing"

        uploaded_artifacts = [
            create_remote_artifact(id=10, workspace=workspace_name),
        ]

        debusine_mock.upload_artifact.side_effect = uploaded_artifacts

        # self.task.workspace_name is set by the Worker
        # and is the workspace that downloads the artifact
        # containing the files needed for debdiff
        self.task.workspace_name = workspace_name

        # The worker set self.task.work_request_id of the task
        work_request_id = 147
        self.task.work_request_id = work_request_id

        self.task.upload_artifacts(
            exec_dir, execution_result=WorkRequestResults.SUCCESS
        )

        # Debusine Mock upload_artifact expected calls
        upload_artifact_calls = []
        debdiff_artifact = DebDiffArtifact.create(
            debdiff_output=debdiff_output,
            original="",
            new="",
            label="debdiff A → B",
        )

        upload_artifact_calls.append(
            call(
                debdiff_artifact,
                workspace=workspace_name,
                work_request=work_request_id,
            )
        )

        # Debusine mock relation_create expected calls
        relation_create_calls = []

        for uploaded_artifact, source_artifact_id in itertools.product(
            uploaded_artifacts, self.task._source_artifacts_ids
        ):
            relation_create_calls.append(
                call(uploaded_artifact.id, source_artifact_id, "relates-to")
            )

        for uploaded_artifact, binary_artifact_id in itertools.product(
            uploaded_artifacts, [binary_artifact_id_1, binary_artifact_id_2]
        ):
            relation_create_calls.append(
                call(uploaded_artifact.id, binary_artifact_id, "relates-to")
            )

        # Assert that the artifacts were uploaded and relations created
        debusine_mock.upload_artifact.assert_has_calls(upload_artifact_calls)
        debusine_mock.relation_create.assert_has_calls(relation_create_calls)

    def test_execute(self) -> None:
        """Test full (mocked) execution."""
        self.configure_task(override={"input": {'source_artifacts': [1, 2]}})
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = DebDiffDynamicData(
            environment_id=1,
            input_source_artifacts_ids=[1, 2],
            parameter_summary="A → B",
        )
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponseArtifact(
            artifact=1
        )
        debusine_mock.download_artifact.return_value = True
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=2, workspace=self.task.workspace_name
        )

        self.assertTrue(self.task.fetch_input(download_directory))

        self.assertEqual(len(self.task._original_dirs), 1)
        self.assertEqual(len(self.task._new_targets_dirs), 1)

        f1_contents = "Source: foo"
        (file1 := self.task._original_dirs[0] / "file1.dsc").write_text(
            f1_contents
        )

        f2_contents = "Source: bar"
        (file2 := self.task._new_targets_dirs[0] / "file2.dsc").write_text(
            f2_contents
        )

        self.patch_prepare_executor_instance()

        self.assertTrue(self.task.configure_for_execution(download_directory))

        self.assertEqual(self.task._original_targets, [file1])
        self.assertEqual(self.task._new_targets, [file2])
        self.assertEqual(
            self.task._cmdline(),
            [
                "debdiff",
                str(file1),
                str(file2),
            ],
        )

        # Create file that will be attached when uploading the artifacts
        debdiff_output = download_directory / self.task.CAPTURE_OUTPUT_FILENAME
        debdiff_output.write_text("1c1\n< Source: foo\n---\n> Source: bar\n")

        self.task.upload_artifacts(
            download_directory, execution_result=WorkRequestResults.SUCCESS
        )

        debusine_mock.upload_artifact.assert_called_once_with(
            DebDiffArtifact.create(
                debdiff_output=download_directory
                / DebDiff.CAPTURE_OUTPUT_FILENAME,
                original=file1.name,
                new=file2.name,
                label="debdiff A → B",
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )
        assert debusine_mock.relation_create.call_count == 2
        debusine_mock.relation_create.assert_has_calls(
            [call(2, 1, "relates-to"), call(2, 2, "relates-to")]
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Paginator extensions for debusine."""

from debusine.web.views.table.columns import (
    Column,
    DateTimeColumn,
    NumberColumn,
    StringColumn,
)
from debusine.web.views.table.filters import (
    FilterField,
    FilterSelectOne,
    FilterText,
    FilterToggle,
)
from debusine.web.views.table.ordering import (
    DateTimeOrdering,
    NumberOrdering,
    OrderBys,
    Ordering,
    Sort,
    StringOrdering,
)
from debusine.web.views.table.paginator import Paginator
from debusine.web.views.table.table import Table
from debusine.web.views.table.views import TableMixin

__all__ = [
    "Column",
    "DateTimeColumn",
    "DateTimeOrdering",
    "FilterField",
    "FilterSelectOne",
    "FilterText",
    "FilterToggle",
    "NumberColumn",
    "NumberOrdering",
    "OrderBys",
    "Ordering",
    "Paginator",
    "Sort",
    "StringColumn",
    "StringOrdering",
    "Table",
    "TableMixin",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the models."""

import datetime as dt
from pathlib import Path
from typing import override
from unittest import mock

import django.db.utils
from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied, ValidationError
from django.test import RequestFactory
from django.utils import timezone

from debusine.artifacts.local_artifact import WorkRequestDebugLogs
from debusine.artifacts.models import ArtifactCategory, EmptyArtifactData
from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    File,
    FileInArtifact,
    FileUpload,
    Scope,
    Workspace,
)
from debusine.db.models.permissions import Allow
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground.scenarios import DefaultContext
from debusine.test.django import (
    AllowAll,
    DenyAll,
    ListFilter,
    TestCase,
    override_permission,
)
from debusine.test.test_utils import data_generator


class ArtifactManagerTests(TestCase):
    """Tests for the ArtifactManager class."""

    scenario = DefaultContext()

    def test_create_from_local_artifact(self) -> None:
        """create_from_local_artifact() creates a matching Artifact."""
        temp_dir = self.create_temporary_directory()
        paths = [temp_dir / "1.log", temp_dir / "2.log"]
        paths[0].write_text("1\n")
        paths[1].write_text("2\n")
        local_artifact = WorkRequestDebugLogs.create(files=paths)
        work_request = self.playground.create_worker_task()

        with context.disable_permission_checks():
            artifact = Artifact.objects.create_from_local_artifact(
                local_artifact,
                self.scenario.workspace,
                created_by_work_request=work_request,
            )

        self.assertEqual(artifact.category, local_artifact.category)
        self.assertEqual(artifact.workspace, self.scenario.workspace)
        self.assertEqual(artifact.data, local_artifact.data.model_dump())
        self.assertEqual(artifact.created_by_work_request, work_request)
        self.assertEqual(artifact.fileinartifact_set.count(), 2)
        for file_in_artifact, path in zip(
            artifact.fileinartifact_set.order_by("id"), paths
        ):
            self.assertEqual(file_in_artifact.path, path.name)
            fileobj = file_in_artifact.file
            file_backend = self.scenario.scope.download_file_backend(fileobj)
            with file_backend.get_stream(fileobj) as file:
                self.assertEqual(file.read(), path.read_bytes())

    @context.disable_permission_checks()
    def test_not_expired_return_not_expired_artifacts(self) -> None:
        """not_expired() return only the not expired artifacts."""
        artifact_1 = Artifact.objects.create(
            workspace=self.scenario.workspace,
            expiration_delay=dt.timedelta(days=1),
        )
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()
        artifact_2 = Artifact.objects.create(
            workspace=self.scenario.workspace,
            expiration_delay=dt.timedelta(days=1),
        )
        artifact_3 = Artifact.objects.create(
            workspace=self.scenario.workspace, expiration_delay=dt.timedelta(0)
        )

        self.assertQuerySetEqual(
            Artifact.objects.not_expired(timezone.now()),
            {artifact_2, artifact_3},
            ordered=False,
        )

    @context.disable_permission_checks()
    def test_expired_return_expired_artifacts(self) -> None:
        """expired() return only the expired artifacts."""
        artifact_1 = Artifact.objects.create(
            workspace=self.scenario.workspace,
            expiration_delay=dt.timedelta(days=1),
        )
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()

        Artifact.objects.create(
            workspace=self.scenario.workspace,
            expiration_delay=dt.timedelta(days=1),
        )
        Artifact.objects.create(
            workspace=self.scenario.workspace, expiration_delay=dt.timedelta(0)
        )

        self.assertQuerySetEqual(
            Artifact.objects.expired(timezone.now()), {artifact_1}
        )

    def test_in_current_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter."""
        artifact, _ = self.playground.create_artifact(
            workspace=self.scenario.workspace
        )

        with context.local():
            context.set_permission_context_scoped(
                self.scenario.scope, PermissionContext.for_ui_anonymous()
            )
            self.assertQuerySetEqual(
                Artifact.objects.in_current_scope(),
                [artifact],
            )

        scope1 = Scope.objects.create(name="Scope1")
        with context.local():
            context.set_permission_context_scoped(
                scope1, PermissionContext.for_ui_anonymous()
            )
            self.assertQuerySetEqual(
                Artifact.objects.in_current_scope(),
                [],
            )

    def test_in_current_scope_no_context_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter without scope set."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "scope is not set"
        ):
            Artifact.objects.in_current_scope()

    def test_in_current_workspace(self) -> None:
        """Test the in_current_workspace() QuerySet filter."""
        artifact, _ = self.playground.create_artifact(
            workspace=self.scenario.workspace
        )
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )

        with context.local():
            self.scenario.workspace.set_current()
            self.assertQuerySetEqual(
                Artifact.objects.in_current_workspace(),
                [artifact],
            )

        workspace1 = self.playground.create_workspace(name="other", public=True)
        with context.local():
            workspace1.set_current()
            self.assertQuerySetEqual(
                Artifact.objects.in_current_workspace(),
                [],
            )

    def test_in_current_workspace_no_context_workspace(self) -> None:
        """Test in_current_workspace() without workspace set."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "workspace is not set"
        ):
            Artifact.objects.in_current_workspace()

    def test_visible_to_current_workspace(self) -> None:
        """Test the visible_to_current_workspace() QuerySet filter."""
        artifact, _ = self.playground.create_artifact(
            workspace=self.scenario.workspace
        )

        with context.local():
            context.set_permission_context_scoped(
                self.scenario.scope, self.scenario.context_ui()
            )

            with self.subTest(workspace="direct"), context.local():
                self.scenario.workspace.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [artifact],
                )

            workspace1 = self.playground.create_workspace(
                name="other", public=True
            )
            with (
                self.subTest(workspace="other", inherits=False),
                context.local(),
            ):
                workspace1.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [],
                )
            workspace1.set_inheritance([self.scenario.workspace])
            with (
                self.subTest(workspace="other", inherits=True),
                context.local(),
            ):
                workspace1.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [artifact],
                )

            grandchild_workspace = self.playground.create_workspace(
                name="grandchild", public=True
            )
            grandchild_workspace.set_inheritance([workspace1])
            with self.subTest(workspace="grandchild"), context.local():
                grandchild_workspace.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [artifact],
                )

            workspace2 = self.playground.create_workspace(
                name="multiple-children", public=True
            )
            workspace2_artifact, _ = self.playground.create_artifact(
                workspace=workspace2
            )
            for i in range(2):
                child_workspace = self.playground.create_workspace(
                    name=f"child-{i}", public=True
                )
                child_workspace.set_inheritance([workspace2])
            with self.subTest(workspace="multiple-children"), context.local():
                workspace2.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [workspace2_artifact],
                )

        scope1 = Scope.objects.create(name="Scope1")
        workspace2 = self.playground.create_workspace(
            name="workspace2", scope=scope1, public=True
        )

        with context.local():
            context.set_permission_context_scoped(
                scope1, self.scenario.context_ui()
            )
            workspace2.set_current()
            with self.subTest(scope="other", inherits=False):
                self.assertQuerySetEqual(
                    Artifact.objects.in_current_scope(),
                    [],
                )
            workspace2.set_inheritance([self.scenario.workspace])
            with self.subTest(scope="other", inherits=True):
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [artifact],
                )

    def test_visible_to_current_workspace_work_request_token(self) -> None:
        private_workspace = self.playground.create_workspace(
            name="private", public=False
        )
        work_request = self.playground.create_worker_task(
            assign_new_worker=True, mark_running=True
        )
        self.playground.create_group_role(
            private_workspace,
            Workspace.Roles.VIEWER,
            users=[work_request.created_by],
        )
        token = self.playground.create_work_request_token(work_request)

        public_artifact, _ = self.playground.create_artifact(
            workspace=self.scenario.workspace
        )
        private_artifact, _ = self.playground.create_artifact(
            workspace=private_workspace
        )

        with context.local():
            context.set_permission_context_scoped(
                self.scenario.scope,
                PermissionContext.for_work_request(token=token),
            )

            with self.subTest(workspace="public"), context.local():
                self.scenario.workspace.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [public_artifact],
                )

            with (
                self.subTest(workspace="private", inherits=False),
                context.local(),
            ):
                private_workspace.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [private_artifact],
                )

            private_workspace.set_inheritance([self.scenario.workspace])
            with (
                self.subTest(workspace="private", inherits=True),
                context.local(),
            ):
                private_workspace.set_current()
                self.assertQuerySetEqual(
                    Artifact.objects.visible_to_current_workspace(),
                    [public_artifact, private_artifact],
                    ordered=False,
                )


class ArtifactTests(TestCase):
    """Tests for the Artifact class."""

    scenario = DefaultContext()

    def assertPermissionDelegatesToWorkspace(
        self, artifact: Artifact, perm_name: str, workspace_perm_name: str
    ) -> None:
        pred_check = getattr(artifact, perm_name)
        allows_anonymous = pred_check.anonymous == Allow.PASS

        with override_permission(Workspace, workspace_perm_name, AllowAll):
            if allows_anonymous:
                self.assertPermission(
                    perm_name,
                    users=(AnonymousUser(), self.scenario.user),
                    allowed=artifact,
                )
            else:
                self.assertPermission(
                    perm_name,
                    users=AnonymousUser(),
                    denied=artifact,
                )
                self.assertPermission(
                    perm_name,
                    users=self.scenario.user,
                    allowed=artifact,
                )
        with override_permission(Workspace, workspace_perm_name, DenyAll):
            self.assertPermission(
                perm_name,
                users=(AnonymousUser(), self.scenario.user),
                denied=artifact,
            )

    def test_save_permissions(self) -> None:
        """Test permission checks on save."""
        for user, username in (
            (None, "AnonymousUser"),
            (self.scenario.user, str(self.scenario.user)),
        ):
            with (
                self.subTest(user=str(user)),
                context.local(),
                override_permission(Workspace, "can_create_artifacts", DenyAll),
                self.assertRaisesRegex(
                    PermissionDenied,
                    (
                        rf"{username} cannot create artifacts in "
                        rf"{self.scenario.workspace}"
                    ),
                ),
            ):
                if user:
                    context.set_permission_context_scoped(
                        self.scenario.scope, PermissionContext.for_ui_user(user)
                    )
                else:
                    context.set_permission_context_scoped(
                        self.scenario.scope,
                        PermissionContext.for_ui_anonymous(),
                    )
                Artifact.objects.create(workspace=self.scenario.workspace)

    @context.disable_permission_checks()
    def test_default_values_of_fields(self) -> None:
        """Define the fields of the models and test their default values."""
        artifact = Artifact()

        self.assertEqual(artifact.category, "")

        artifact.category = "sample-type"

        self.assertEqual(artifact.data, {})

        artifact.workspace = self.playground.create_workspace(name="test")

        artifact.clean_fields()

        date_time_before_save = timezone.now()

        artifact.save()

        # date_time_before_save <= artifact.created_at <= timezone.now()
        self.assertLessEqual(date_time_before_save, artifact.created_at)
        self.assertLessEqual(artifact.created_at, timezone.now())

        # By default, the expiration_delay is None
        self.assertIsNone(artifact.expiration_delay)

        # By default, the created_by_work_request is None
        self.assertIsNone(artifact.created_by_work_request)

        # By default, the created_by is None
        self.assertIsNone(artifact.created_by)

    @context.disable_permission_checks()
    def test_expired(self) -> None:
        """Test Artifact.expired() method."""
        artifact = Artifact()

        artifact.created_at = timezone.now() - dt.timedelta(days=2)
        artifact.expiration_delay = dt.timedelta(days=1)
        self.assertTrue(artifact.expired(timezone.now()))

        artifact.expiration_delay = dt.timedelta(0)
        self.assertFalse(artifact.expired(timezone.now()))

        artifact.created_at = timezone.now()
        artifact.expiration_delay = dt.timedelta(days=1)
        self.assertFalse(artifact.expired(timezone.now()))

    @context.disable_permission_checks()
    def test_default_expiration(self) -> None:
        """Test workspace's default_expiration_delay."""
        self.assertEqual(
            self.scenario.workspace.default_expiration_delay, dt.timedelta(0)
        )
        artifact = Artifact.objects.create(workspace=self.scenario.workspace)
        self.assertIsNone(artifact.expiration_delay)

        expiring_ws = self.playground.create_workspace(name="test")
        expiring_ws.default_expiration_delay = dt.timedelta(days=7)
        test_creation_time = timezone.now()
        artifact = Artifact.objects.create(workspace=expiring_ws)
        delta = expiring_ws.default_expiration_delay
        assert artifact.expire_at is not None
        self.assertEqual(
            (artifact.expire_at - test_creation_time).days, delta.days
        )

    @context.disable_permission_checks()
    def test_str(self) -> None:
        """Test for Artifact.__str__."""
        artifact = Artifact.objects.create(workspace=self.scenario.workspace)
        self.assertEqual(
            artifact.__str__(),
            f"Id: {artifact.id} "
            f"Category: {artifact.category} "
            f"Workspace: {artifact.workspace.id}",
        )

    def test_ui(self) -> None:
        from debusine.web.views.ui.artifacts import ArtifactUI

        artifact, _ = self.playground.create_artifact()
        request = RequestFactory().get("/")
        ui = artifact.ui(request)
        self.assertIsInstance(ui, ArtifactUI)

    @context.disable_permission_checks()
    def test_get_absolute_url(self) -> None:
        """Test for Artifact.get_absolute_url."""
        artifact = Artifact.objects.create(workspace=self.scenario.workspace)
        self.assertEqual(
            artifact.get_absolute_url(),
            f"/{self.scenario.scope.name}/{self.scenario.workspace.name}"
            f"/artifact/{artifact.pk}/",
        )

    @context.disable_permission_checks()
    def test_get_absolute_url_different_scope(self) -> None:
        """Artifact.get_absolute_url works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        artifact = Artifact.objects.create(workspace=other_workspace)
        self.assertEqual(
            artifact.get_absolute_url(),
            f"/{other_scope.name}/{other_workspace.name}"
            f"/artifact/{artifact.pk}/",
        )

    @context.disable_permission_checks()
    def test_get_absolute_url_download(self) -> None:
        """Test for Artifact.get_absolute_url_download."""
        artifact = Artifact.objects.create(workspace=self.scenario.workspace)
        self.assertEqual(
            artifact.get_absolute_url_download(),
            f"/{self.scenario.scope.name}/{self.scenario.workspace.name}"
            f"/artifact/{artifact.pk}/download/",
        )

    @context.disable_permission_checks()
    def test_get_absolute_url_download_different_scope(self) -> None:
        """Artifact.get_absolute_url_download works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        artifact = Artifact.objects.create(workspace=other_workspace)
        self.assertEqual(
            artifact.get_absolute_url_download(),
            f"/{other_scope.name}/{other_workspace.name}"
            f"/artifact/{artifact.pk}/download/",
        )

    def test_data_validation_error(self) -> None:
        """Assert that ValidationError is raised if data is not valid."""
        with context.disable_permission_checks():
            artifact = Artifact.objects.create(
                category=ArtifactCategory.WORK_REQUEST_DEBUG_LOGS,
                workspace=self.scenario.workspace,
                data=[],
            )
        with self.assertRaises(ValidationError) as raised:
            artifact.full_clean()
        self.assertEqual(
            raised.exception.message_dict,
            {"data": ["data must be a dictionary"]},
        )

        artifact.category = "missing:does-not-exist"
        artifact.data = {}
        with self.assertRaises(ValidationError) as raised:
            artifact.full_clean()
        self.assertEqual(
            raised.exception.message_dict,
            {"category": ["missing:does-not-exist: invalid artifact category"]},
        )

        artifact.category = ArtifactCategory.PACKAGE_BUILD_LOG
        with self.assertRaises(ValidationError) as raised:
            artifact.full_clean()
        messages = raised.exception.message_dict
        self.assertCountEqual(messages.keys(), ["data"])
        self.assertRegex(
            messages["data"][0],
            r"invalid artifact data:"
            r" 3 validation errors for DebianPackageBuildLog\n",
        )

    @context.disable_permission_checks()
    def test_get_label_from_data(self) -> None:
        """Test getting artifact label from its data."""
        artifact = Artifact.objects.create(
            category=ArtifactCategory.SOURCE_PACKAGE,
            workspace=self.scenario.workspace,
            data={
                "name": "test",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
        )
        self.assertEqual(artifact.get_label(), "test_1.0-1")

    @context.disable_permission_checks()
    def test_get_label_from_category(self) -> None:
        """Test defaults for artifact label."""
        artifact = Artifact.objects.create(
            category=ArtifactCategory.WORK_REQUEST_DEBUG_LOGS,
            workspace=self.scenario.workspace,
            data={},
        )
        self.assertEqual(
            artifact.get_label(), "debusine:work-request-debug-logs"
        )

    @context.disable_permission_checks()
    def test_get_label_from_data_reuse_model(self) -> None:
        """Test getting artifact label reusing the model instance."""
        artifact = Artifact.objects.create(
            category=ArtifactCategory.SOURCE_PACKAGE,
            workspace=self.scenario.workspace,
            data={
                "name": "test",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
        )
        self.assertEqual(
            artifact.get_label(EmptyArtifactData()), "debian:source-package"
        )

    def test_can_display_delegate_to_workspace(self) -> None:
        """Test the can_display predicate on public workspaces."""
        artifact, _ = self.playground.create_artifact()
        self.assertPermissionDelegatesToWorkspace(
            artifact, "can_display", "can_display"
        )

    def test_can_display_work_request_created_artifacts(self) -> None:
        """Test the can_display predicate on public workspaces."""
        work_request = self.playground.create_worker_task(mark_running=True)
        token = self.playground.create_work_request_token(work_request)

        artifact, _ = self.playground.create_artifact()
        artifact.created_by_work_request = work_request
        artifact.save()

        other, _ = self.playground.create_artifact()

        with override_permission(Workspace, "can_display", AllowAll):
            self.assertPermission(
                "can_display",
                users=[AnonymousUser(), self.scenario.user],
                allowed=artifact,
                denied=other,
                token=token,
            )
        with override_permission(Workspace, "can_display", DenyAll):
            self.assertPermission(
                "can_display",
                users=(AnonymousUser(), self.scenario.user),
                denied=[artifact, other],
            )

    def test_can_display_work_request_input_allowlist(self) -> None:
        """Test the can_display predicate on public workspaces."""
        task_input, _ = self.playground.create_artifact()
        other, _ = self.playground.create_artifact()
        work_request = self.playground.create_worker_task(mark_running=True)
        token = self.playground.create_work_request_token(work_request)
        self.enterContext(
            mock.patch(
                "debusine.tasks.models.BaseDynamicTaskData"
                ".get_input_artifacts_ids",
                return_value=[task_input.pk],
            )
        )
        with override_permission(Workspace, "can_display", AllowAll):
            self.assertPermission(
                "can_display",
                users=[AnonymousUser(), self.scenario.user],
                allowed=task_input,
                denied=other,
                token=token,
            )
        with override_permission(Workspace, "can_display", DenyAll):
            self.assertPermission(
                "can_display",
                users=(AnonymousUser(), self.scenario.user),
                denied=[task_input, other],
            )

    def test_can_add_files_delegate_to_workspace(self) -> None:
        artifact, _ = self.playground.create_artifact()
        self.assertPermissionDelegatesToWorkspace(
            artifact, "can_add_files", "can_create_artifacts"
        )

    def test_can_add_files_work_request(self) -> None:
        work_request = self.playground.create_worker_task(mark_running=True)
        token = self.playground.create_work_request_token(work_request)

        artifact, _ = self.playground.create_artifact()
        artifact.created_by_work_request = work_request
        artifact.save()

        other, _ = self.playground.create_artifact()

        with override_permission(Workspace, "can_create_artifacts", AllowAll):
            self.assertPermission(
                "can_add_files",
                users=AnonymousUser(),
                denied=[artifact, other],
            )
            self.assertPermission(
                "can_add_files",
                users=[self.scenario.user],
                allowed=artifact,
                denied=other,
                token=token,
            )
        with override_permission(Workspace, "can_create_artifacts", DenyAll):
            self.assertPermission(
                "can_add_files",
                users=(AnonymousUser(), self.scenario.user),
                denied=[artifact, other],
            )

    def test_can_add_relation_delegate_to_workspace(self) -> None:
        artifact, _ = self.playground.create_artifact()
        self.assertPermissionDelegatesToWorkspace(
            artifact, "can_add_relation", "can_create_artifacts"
        )

    def test_can_add_relation_work_request(self) -> None:
        work_request = self.playground.create_worker_task(mark_running=True)
        token = self.playground.create_work_request_token(work_request)

        artifact, _ = self.playground.create_artifact()
        artifact.created_by_work_request = work_request
        artifact.save()

        other, _ = self.playground.create_artifact()

        with override_permission(Workspace, "can_create_artifacts", AllowAll):
            self.assertPermission(
                "can_add_relation",
                users=AnonymousUser(),
                denied=[artifact, other],
            )
            self.assertPermission(
                "can_add_relation",
                users=[self.scenario.user],
                allowed=artifact,
                denied=other,
                token=token,
            )
        with override_permission(Workspace, "can_create_artifacts", DenyAll):
            self.assertPermission(
                "can_add_relation",
                users=(AnonymousUser(), self.scenario.user),
                denied=[artifact, other],
            )


class FileInArtifactTests(TestCase):
    """Tests for the FileInArtifact class."""

    def create_file_in_artifact(
        self,
        fileobj: File | None = None,
        workspace: Workspace | None = None,
        artifact: Artifact | None = None,
        path: str | None = None,
        complete: bool = False,
    ) -> FileInArtifact:
        """Return FileInArtifact to be used in tests."""
        if fileobj is None:
            fileobj = self.playground.create_file()

        if artifact is None:
            artifact, _ = self.playground.create_artifact(workspace=workspace)

        if path is None:
            path = "/usr/bin/test"

        return FileInArtifact.objects.create(
            artifact=artifact,
            file=fileobj,
            path=path,
            complete=complete,
        )

    def test_artifact_path_unique_constraint(self) -> None:
        """Test two FileInArtifact cannot have the same artifact and path."""
        artifact, _ = self.playground.create_artifact()

        file1 = self.playground.create_file(b"contents1")
        file2 = self.playground.create_file(b"contents2")

        self.create_file_in_artifact(
            file1, artifact=artifact, path="/usr/bin/test"
        )

        with self.assertRaisesRegex(
            django.db.utils.IntegrityError,
            "db_fileinartifact_unique_artifact_path",
        ):
            self.create_file_in_artifact(
                file2, artifact=artifact, path="/usr/bin/test"
            )

    def test_str(self) -> None:
        """Test FileInArtifact.__str__."""
        file_in_artifact = self.create_file_in_artifact()

        self.assertEqual(
            file_in_artifact.__str__(),
            f"Id: {file_in_artifact.id} "
            f"Artifact: {file_in_artifact.artifact.id} "
            f"Path: {file_in_artifact.path} "
            f"File: {file_in_artifact.file.id}",
        )

    def test_ui(self) -> None:
        from debusine.web.views.ui.artifacts import FileInArtifactUI

        file_in_artifact = self.create_file_in_artifact()
        request = RequestFactory().get("/")
        ui = file_in_artifact.ui(request)
        self.assertIsInstance(ui, FileInArtifactUI)

    def test_get_absolute_url(self) -> None:
        """Test FileInArtifact.get_absolute_url."""
        file_in_artifact = self.create_file_in_artifact()
        self.assertEqual(
            file_in_artifact.get_absolute_url(),
            f"/debusine/System/artifact"
            f"/{file_in_artifact.artifact.id}/file//usr/bin/test",
        )

    def test_get_absolute_url_different_scope(self) -> None:
        """FileInArtifact.get_absolute_url works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        file_in_artifact = self.create_file_in_artifact(
            workspace=other_workspace
        )
        self.assertEqual(
            file_in_artifact.get_absolute_url(),
            f"/{other_scope.name}/{other_workspace.name}/artifact"
            f"/{file_in_artifact.artifact.id}/file//usr/bin/test",
        )

    def test_get_absolute_url_raw(self) -> None:
        """Test FileInArtifact.get_absolute_url_raw."""
        file_in_artifact = self.create_file_in_artifact()
        self.assertEqual(
            file_in_artifact.get_absolute_url_raw(),
            f"/debusine/System/artifact"
            f"/{file_in_artifact.artifact.id}/raw//usr/bin/test",
        )

    def test_get_absolute_url_raw_different_scope(self) -> None:
        """FileInArtifact.get_absolute_url_raw works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        file_in_artifact = self.create_file_in_artifact(
            workspace=other_workspace
        )
        self.assertEqual(
            file_in_artifact.get_absolute_url_raw(),
            f"/{other_scope.name}/{other_workspace.name}/artifact"
            f"/{file_in_artifact.artifact.id}/raw//usr/bin/test",
        )

    def test_get_absolute_url_download(self) -> None:
        """Test FileInArtifact.get_absolute_url_download."""
        file_in_artifact = self.create_file_in_artifact()
        self.assertEqual(
            file_in_artifact.get_absolute_url_download(),
            f"/debusine/System/artifact"
            f"/{file_in_artifact.artifact.id}/download//usr/bin/test",
        )

    def test_get_absolute_url_download_different_scope(self) -> None:
        """FileInArtifact.get_absolute_url_download works in another scope."""
        other_scope = self.playground.get_or_create_scope(name="other")
        other_workspace = self.playground.create_workspace(scope=other_scope)
        file_in_artifact = self.create_file_in_artifact(
            workspace=other_workspace
        )
        self.assertEqual(
            file_in_artifact.get_absolute_url_download(),
            f"/{other_scope.name}/{other_workspace.name}/artifact"
            f"/{file_in_artifact.artifact.id}/download//usr/bin/test",
        )


class FileUploadTests(TestCase):
    """Tests for FileUpload class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.file_upload = self.playground.create_file_upload()
        self.artifact = self.file_upload.file_in_artifact.artifact
        self.workspace = self.artifact.workspace
        self.file_path = self.file_upload.file_in_artifact.path
        self.file_size = self.file_upload.file_in_artifact.file.size

    def test_current_size_raise_no_file_in_artifact(self) -> None:
        """Test current_size() raise ValueError (FileInArtifact not found)."""
        wrong_path = "no-exist"
        with self.assertRaisesRegex(
            ValueError,
            f'^No FileInArtifact for Artifact {self.artifact.id} '
            f'and path "{wrong_path}"$',
        ):
            FileUpload.current_size(self.artifact, wrong_path)

    @context.disable_permission_checks()
    def test_current_size_raise_no_fileupload_for_file_in_artifact(
        self,
    ) -> None:
        """Test current_size() raise ValueError (FileUpload not found)."""
        artifact = Artifact.objects.create(
            category="test", workspace=self.workspace
        )
        file_in_artifact = FileInArtifact.objects.create(
            artifact=artifact,
            path="something",
            file=self.playground.create_file(),
        )

        with self.assertRaisesRegex(
            ValueError,
            f"^No FileUpload for FileInArtifact {file_in_artifact.id}$",
        ):
            FileUpload.current_size(artifact, "something")

    def test_current_size_return_last_position_received(self) -> None:
        """Test current_size() method return size of the file."""
        write_to_position = 30
        self.file_upload.absolute_file_path().write_bytes(
            next(data_generator(write_to_position))
        )
        self.assertEqual(
            FileUpload.current_size(
                artifact=self.artifact, path_in_artifact=self.file_path
            ),
            write_to_position,
        )

    def test_no_two_fileobj_to_same_path(self) -> None:
        """Test cannot create two FileUpload with same path."""
        data = next(data_generator(self.file_size))
        file = self.playground.create_file(data)
        new_file_in_artifact = FileInArtifact.objects.create(
            artifact=self.artifact, path="README-2", file=file
        )

        with self.assertRaisesRegex(
            django.db.utils.IntegrityError, "db_fileupload_path_key"
        ):
            FileUpload.objects.create(
                file_in_artifact=new_file_in_artifact,
                path=self.file_upload.path,
            )

    def test_delete(self) -> None:
        """Test FileUpload delete() try to unlink file."""
        file_path = self.file_upload.absolute_file_path()
        Path(file_path).write_bytes(next(data_generator(self.file_size)))

        self.assertTrue(file_path.exists())

        with self.captureOnCommitCallbacks(execute=True) as callbacks:
            self.file_upload.delete()

        self.assertFalse(file_path.exists())

        self.assertEqual(len(callbacks), 1)

    def test_delete_file_does_not_exist(self) -> None:
        """Test FileUpload delete() try to unlink file but did not exist."""
        file_path = self.file_upload.absolute_file_path()

        self.assertFalse(Path(file_path).exists())

        with self.captureOnCommitCallbacks(execute=True) as callbacks:
            self.file_upload.delete()

        self.assertEqual(len(callbacks), 1)

    def test_absolute_file_path(self) -> None:
        """Test absolute_file_path() is in DEBUSINE_UPLOAD_DIRECTORY."""
        self.assertEqual(
            self.file_upload.absolute_file_path(),
            Path(settings.DEBUSINE_UPLOAD_DIRECTORY) / self.file_upload.path,
        )

    def test_str(self) -> None:
        """Test __str__."""
        self.assertEqual(self.file_upload.__str__(), f"{self.file_upload.id}")


class ArtifactRelationManagerTests(TestCase):
    """Tests for the ArtifactRelationManager class."""

    scenario = DefaultContext()

    def test_in_current_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter."""
        artifact, _ = self.playground.create_artifact(
            workspace=self.scenario.workspace
        )
        target, _ = self.playground.create_artifact(
            workspace=self.scenario.workspace
        )
        relation = self.playground.create_artifact_relation(
            artifact=artifact, target=target
        )

        with context.local():
            self.scenario.set_current()
            self.assertQuerySetEqual(
                ArtifactRelation.objects.in_current_scope(), [relation]
            )

        scope1 = Scope.objects.create(name="Scope1")
        with context.local():
            context.set_permission_context_scoped(
                scope1, PermissionContext.for_ui_anonymous()
            )
            self.assertQuerySetEqual(
                ArtifactRelation.objects.in_current_scope(), []
            )

    def test_in_current_scope_no_context_scope(self) -> None:
        """Test the in_current_scope() QuerySet filter without scope set."""
        with self.assertRaisesRegex(
            ContextConsistencyError, "scope is not set"
        ):
            ArtifactRelation.objects.in_current_scope()


class ArtifactRelationTests(TestCase):
    """Implement ArtifactRelation tests."""

    scenario = DefaultContext()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.artifact_1, _ = self.playground.create_artifact()
        self.artifact_2, _ = self.playground.create_artifact(
            workspace=self.playground.create_workspace(name="other-workspace")
        )

    def test_type_not_valid_raise_validation_error(self) -> None:
        """Assert that ValidationError is raised if type is not valid."""
        artifact_relation = ArtifactRelation.objects.create(
            artifact=self.artifact_1,
            target=self.artifact_2,
            type="wrong-type",
        )
        with self.assertRaises(ValidationError):
            artifact_relation.full_clean()

    def test_str(self) -> None:
        """Assert __str__ return expected value."""
        artifact_relation = self.playground.create_artifact_relation(
            artifact=self.artifact_1, target=self.artifact_2
        )
        self.assertEqual(
            str(artifact_relation),
            f"{self.artifact_1.id} {ArtifactRelation.Relations.RELATES_TO} "
            f"{self.artifact_2.id}",
        )

    def test_can_display_delegate_to_workspace(self) -> None:
        """Test the can_display predicate on public workspaces."""
        artifact_relation = self.playground.create_artifact_relation(
            artifact=self.artifact_1, target=self.artifact_2
        )
        with override_permission(Workspace, "can_display", AllowAll):
            self.assertPermission(
                "can_display",
                users=(AnonymousUser(), self.scenario.user),
                allowed=artifact_relation,
            )
        for allowed_artifact in (self.artifact_1, self.artifact_2):
            with override_permission(
                Workspace,
                "can_display",
                ListFilter,
                include=[allowed_artifact.workspace],
            ):
                self.assertPermission(
                    "can_display",
                    users=(AnonymousUser(), self.scenario.user),
                    denied=artifact_relation,
                )
        with override_permission(Workspace, "can_display", DenyAll):
            self.assertPermission(
                "can_display",
                users=(AnonymousUser(), self.scenario.user),
                denied=artifact_relation,
            )

    def test_can_display_both_artifacts_same_wr(self) -> None:
        wr = self.playground.create_worker_task()
        self.artifact_1.created_by_work_request = wr
        self.artifact_1.save()
        self.artifact_2.workspace = self.scenario.workspace
        self.artifact_2.created_by_work_request = wr
        self.artifact_2.save()
        artifact_relation = self.playground.create_artifact_relation(
            artifact=self.artifact_1, target=self.artifact_2
        )
        pc = PermissionContext.for_work_request(work_request=wr)
        self.assertTrue(artifact_relation.can_display(pc))
        self.assertQuerySetEqual(
            ArtifactRelation.objects.can_display(pc), [artifact_relation]
        )

    def test_can_display_both_artifacts_as_input(self) -> None:
        wr = self.playground.create_worker_task()
        self.artifact_1.created_by_work_request = wr
        self.artifact_1.save()
        self.artifact_2.workspace = self.scenario.workspace
        self.artifact_2.save()
        self.enterContext(
            mock.patch(
                "debusine.tasks.BaseExternalTask.get_input_artifacts_ids",
                return_value=[self.artifact_1.pk, self.artifact_2.pk],
            )
        )
        artifact_relation = self.playground.create_artifact_relation(
            artifact=self.artifact_1, target=self.artifact_2
        )
        pc = PermissionContext.for_work_request(work_request=wr)
        self.assertTrue(artifact_relation.can_display(pc))
        self.assertQuerySetEqual(
            ArtifactRelation.objects.can_display(pc), [artifact_relation]
        )

    def test_can_delete(self) -> None:
        """Test the can_delete predicate."""
        artifact_relation = self.playground.create_artifact_relation(
            artifact=self.artifact_1, target=self.artifact_2
        )
        # Requires OWNER access in the artifact's workspace:
        with (
            self.subTest(workspace_owner=False, can_display=True),
            override_permission(Workspace, "can_display", DenyAll),
        ):
            self.assertPermission(
                "can_delete",
                users=(AnonymousUser(), self.scenario.user),
                denied=artifact_relation,
            )

        self.playground.create_group_role(
            self.artifact_1.workspace,
            Workspace.Roles.OWNER,
            users=[self.scenario.user],
        )

        # Also requires can_display
        with self.subTest(can_display=False):
            self.assertPermission(
                "can_delete",
                users=(AnonymousUser(), self.scenario.user),
                denied=artifact_relation,
            )
        with (
            self.subTest(can_display=True),
            override_permission(Workspace, "can_display", AllowAll),
        ):
            self.assertPermission(
                "can_delete",
                users=AnonymousUser(),
                denied=artifact_relation,
            )
            self.assertPermission(
                "can_delete",
                users=self.scenario.user,
                allowed=artifact_relation,
            )



# Generated by Django 4.2.20 on 2025-05-19 23:16

import django.db.models.fields.json
import django.utils.timezone
from django.db import migrations, models

import debusine.artifacts.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0006_fileinartifact_content_type'),
    ]
    replaces = [
        ('db', '0148_alter_collectionitem_created_at_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionitem',
            name='created_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddIndex(
            model_name='collectionitem',
            index=models.Index(
                models.F('parent_collection'),
                django.db.models.fields.json.KeyTextTransform(
                    'component', 'data'
                ),
                models.F('created_at'),
                models.F('removed_at'),
                condition=models.Q(
                    (
                        'category',
                        debusine.artifacts.models.ArtifactCategory[
                            'SOURCE_PACKAGE'
                        ],
                    ),
                    ('child_type', 'a'),
                    (
                        'parent_category',
                        debusine.artifacts.models.CollectionCategory['SUITE'],
                    ),
                ),
                name='db_ci_suite_all_sources_idx',
            ),
        ),
        migrations.AddIndex(
            model_name='collectionitem',
            index=models.Index(
                models.F('parent_collection'),
                django.db.models.fields.json.KeyTextTransform(
                    'component', 'data'
                ),
                django.db.models.fields.json.KeyTextTransform(
                    'architecture', 'data'
                ),
                models.F('created_at'),
                models.F('removed_at'),
                condition=models.Q(
                    (
                        'category',
                        debusine.artifacts.models.ArtifactCategory[
                            'BINARY_PACKAGE'
                        ],
                    ),
                    ('child_type', 'a'),
                    (
                        'parent_category',
                        debusine.artifacts.models.CollectionCategory['SUITE'],
                    ),
                ),
                name='db_ci_suite_all_binaries_idx',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Workflow to coordinate metadata updates for all archives in a workspace."""

import logging
from typing import override

from django.db.models import Exists, F, Max, Q, QuerySet
from django.db.models.sql.constants import LOUTER
from django_cte import CTE, with_cte

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    TaskTypes,
)
from debusine.assets import KeyPurpose
from debusine.assets.models import AssetCategory, SigningKeyData
from debusine.db.models import (
    AssetUsage,
    Collection,
    CollectionItem,
    WorkRequest,
    Workspace,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.server.workflows import Workflow, WorkflowValidationError
from debusine.server.workflows.models import (
    UpdateSuiteData,
    UpdateSuitesData,
    WorkRequestWorkflowData,
)
from debusine.signing.tasks.models import GenerateKeyData

logger = logging.getLogger(__name__)


class UpdateSuitesWorkflow(Workflow[UpdateSuitesData]):
    """Coordinate metadata updates for all suites in the workspace."""

    TASK_NAME = "update_suites"

    @override
    def validate_input(self) -> None:
        """Thorough validation of input data."""
        # Only a workspace owner may run this workflow, since that implies
        # the SIGNER role on newly-created signing keys.
        if not self.workspace.has_role(
            PermissionContext.for_work_request(work_request=self.work_request),
            Workspace.Roles.OWNER,
        ):
            raise WorkflowValidationError(
                'Only a workspace owner may run the "update_suites" workflow'
            )

    def get_archive(self) -> Collection:
        """
        Return the workspace's archive.

        This deliberately doesn't follow the workspace's inheritance chain.
        """
        return self.workspace.collections.get(
            category=CollectionCategory.ARCHIVE, name="_"
        )

    def _find_stale_suites(self) -> QuerySet[Collection]:
        """
        Return all stale suites in this workspace.

        For this purpose, a suite is "stale" if any source or binary
        packages in it have been changed (collection items created or
        removed) since its most recent ``Release`` file was generated.
        """
        suites: QuerySet[Collection]
        if self.data.force_basic_indexes:
            suites = Collection.objects.all()
        else:
            # Use a couple of CTEs to ensure that PostgreSQL uses a
            # reasonable index to find index:Release items, and that it only
            # does the work of figuring out the latest Release creation time
            # once.

            # Build a virtual table of previously-generated Release indexes.
            release_item = CTE(
                CollectionItem.objects.filter(
                    parent_category=CollectionCategory.SUITE,
                    child_type=CollectionItem.Types.ARTIFACT,
                    category=ArtifactCategory.REPOSITORY_INDEX,
                    name="index:Release",
                ),
                name="release_item",
            )
            # Annotate collections with the most recent Release index
            # creation time in each one.
            collection_generated = CTE(
                release_item.join(
                    Collection,
                    id=release_item.col.parent_collection_id,
                    _join_type=LOUTER,
                )
                .values("id")
                .annotate(latest_created=Max(release_item.col.created_at)),
                name="collection_generated",
            )
            # Find suite names where collection items have been created or
            # removed more recently than their most recent Release index.
            suites = with_cte(
                release_item,
                collection_generated,
                select=(
                    collection_generated.join(
                        Collection, id=collection_generated.col.id
                    )
                    .annotate(
                        changed=Exists(
                            CollectionItem.objects.filter(
                                parent_collection=collection_generated.col.id
                            )
                            .annotate(
                                latest_created=(
                                    collection_generated.col.latest_created
                                )
                            )
                            .filter(
                                Q(latest_created__isnull=True)
                                | Q(created_at__gt=F("latest_created"))
                                | Q(removed_at__gt=F("latest_created"))
                            )
                        ),
                    )
                    .filter(changed=True)
                ),
            )

        if self.data.only_suites is not None:
            suites = suites.filter(name__in=self.data.only_suites)

        return suites.filter(
            workspace=self.workspace,
            id__in=Collection.objects.exported_suites(),
        ).order_by("name")

    def _populate_generate_key(self) -> WorkRequest:
        generate_key, _ = self.work_request_ensure_child_signing(
            task_name="generatekey",
            task_data=GenerateKeyData(
                purpose=KeyPurpose.OPENPGP_REPOSITORY,
                description=f"Archive signing key for {self.workspace}",
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"Generate signing key for {self.workspace}",
                step="generate-signing-key",
            ),
        )
        generated_key, _ = self.work_request_ensure_child_internal(
            task_name="workflow",
            workflow_data=WorkRequestWorkflowData(
                display_name="Record generated key", step="generated-key"
            ),
        )
        generated_key.add_dependency(generate_key)
        self.orchestrate_child(generated_key)
        return generated_key

    def _populate_main(self, generated_key: WorkRequest | None) -> None:
        for suite in self._find_stale_suites():
            update_suite, _ = self.work_request_ensure_child_workflow(
                task_name="update_suite",
                task_data=UpdateSuiteData(suite_collection=suite.id),
                workflow_data=WorkRequestWorkflowData(
                    display_name=f"Update {suite.name}",
                    step=f"update-{suite.name}",
                ),
            )
            if generated_key is not None:
                update_suite.add_dependency(generated_key)

        for update_suite in self.work_request.children.filter(
            task_type=TaskTypes.WORKFLOW,
            task_name="update_suite",
            status__in=(
                WorkRequest.Statuses.BLOCKED,
                WorkRequest.Statuses.PENDING,
            ),
        ):
            self.orchestrate_child(update_suite)

    @override
    def populate(self) -> None:
        """Create child work requests for all stale suites."""
        try:
            archive = self.get_archive()
        except Collection.DoesNotExist:
            logger.info(
                "Workspace %s has no archive; not updating suites.",
                self.workspace,
            )
            return

        generated_key: WorkRequest | None = None
        if "signing_keys" not in archive.data:
            generated_key = self._populate_generate_key()
        self._populate_main(generated_key)

    def callback_generated_key(self) -> bool:
        """Configure the generated key."""
        # Configure the archive to use the new key.
        archive = self.get_archive()
        assert "signing_keys" not in archive.data
        signing_key = self.work_request.children.get(
            task_type=TaskTypes.SIGNING, task_name="generatekey"
        ).asset_set.get(category=AssetCategory.SIGNING_KEY)
        AssetUsage.objects.get_or_create(
            asset=signing_key, workspace=self.workspace
        )
        signing_key_data = signing_key.data_model
        assert isinstance(signing_key_data, SigningKeyData)
        archive.data["signing_keys"] = [signing_key_data.fingerprint]
        archive.save()
        return True



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to create a notification channel."""

import argparse
import sys
from typing import Any, NoReturn, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser

from debusine.db.models import NotificationChannel
from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to create an enabled token."""

    help = "Create a new notification channel"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        methods = dict(NotificationChannel.Methods.choices).keys()
        parser.add_argument("name", help="Notification channel name")
        parser.add_argument(
            "method", help="Notification method", choices=methods
        )
        parser.add_argument(
            "--data",
            type=argparse.FileType("r"),
            help="File path (or - for stdin) to read the data for "
            "the notification. YAML format. Defaults to stdin.",
            default="-",
        )

    @override
    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        if options["data"] != sys.stdin:
            options["data"].close()

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        name = options["name"]
        method = options["method"]
        data = self.parse_yaml_data(options["data"].read())

        try:
            channel = NotificationChannel.objects.create(
                name=name, method=method, data=data
            )
            channel.full_clean()
            channel.save()
        except ValidationError as exc:
            raise CommandError(
                "Error creating notification channel: "
                + "\n".join(exc.messages),
                returncode=3,
            )

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Middleware for updating Worker.last_seen_at."""

from collections.abc import Awaitable, Callable, Iterable
from typing import TYPE_CHECKING

from asgiref.typing import ASGIReceiveCallable, ASGISendCallable
from channels.db import database_sync_to_async
from django.utils import timezone

from debusine.db.context import context
from debusine.db.models import Token, WorkRequest
from debusine.db.permissioncontext import PermissionContext
from debusine.server.consumers import ChannelScopeWithToken

if TYPE_CHECKING:
    from channels.consumer import _ChannelScope


class TokenLastSeenAtMiddlewareChannels:
    """
    Save the Token in the scope and update Worker.last_seen_at.

    Token available in the scope helps to avoid double lookups.
    """

    def __init__(
        self,
        app: Callable[
            [ChannelScopeWithToken, ASGIReceiveCallable, ASGISendCallable],
            Awaitable[None],
        ],
    ) -> None:
        """Middleware API entry point."""
        self.app = app

    @staticmethod
    def get_token_header(headers: Iterable[tuple[bytes, bytes]]) -> str | None:
        """Return the token key from the headers or None."""
        for header in headers:
            if header[0].lower() == b"token":
                return header[1].decode("utf-8")

        return None

    def init_context_with_work_request_token(self, token: Token) -> None:
        """Initialize the context from a work request token."""
        # This is a sync operation, as it will do a database lookup to precache
        # scope and workspace roles
        assert token.work_request is not None
        context.set_permission_context_scoped(
            token.work_request.workspace.scope,
            PermissionContext.for_work_request(token=token),
        )

    async def __call__(
        self,
        scope: "_ChannelScope",
        receive: ASGIReceiveCallable,
        send: ASGISendCallable,
    ) -> None:
        """Middleware entry point."""
        # When middleware is modifying the scope, it should make a copy of the
        # scope object before mutating it.
        # See https://asgi.readthedocs.io/en/latest/specs/main.html#middleware
        scope_with_token = ChannelScopeWithToken(**scope, token=None)

        token_key = self.get_token_header(scope_with_token["headers"])

        if token_key is None:
            # No token was in the request, nothing needs to be done
            scope_with_token["token"] = None
            return await self.app(scope_with_token, receive, send)

        token = await database_sync_to_async(
            Token.objects.select_related(
                "work_request__workspace__scope",
                "work_request__created_by",
            ).get_token_or_none
        )(token_key)

        if token is not None:
            token.last_seen_at = timezone.now()
            await token.asave()
            if token.enabled:
                match token.token_type:
                    case Token.Types.WORKER:
                        context.set_permission_context_scopeless(
                            PermissionContext.for_worker(token)
                        )
                    case Token.Types.WORK_REQUEST:
                        if (
                            token.work_request is not None
                            and token.work_request.status
                            == WorkRequest.Statuses.RUNNING
                        ):
                            await database_sync_to_async(
                                self.init_context_with_work_request_token
                            )(token)
                    case _:
                        pass

        scope_with_token["token"] = token

        return await self.app(scope_with_token, receive, send)
